import logging
from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.bioshake_pb2 import Command, Config

from .driver import BioshakeDriver
import argparse 

class BioShakeServer(ToolServer):
    toolType = "bioshake"

    def __init__(self) -> None:
        super().__init__()
        self.driver: BioshakeDriver

    def _configure(self, config: Config) -> None:
        self.config = config
        if self.driver:
            logging.info("Bioshake driver already exists, reconfiguring...")
            self.driver.disconnect()
        self.driver = BioshakeDriver(port=self.config.com_port)
        self.driver.connect()

    def Grip(self, params: Command.Grip) -> None:
        if not self.driver:
            raise Exception("Bioshake driver not connected")
        self.driver.set_elm_lock_pos()

    def Ungrip(self, params: Command.Ungrip) -> None:
        if not self.driver:
            raise Exception("Bioshake driver not connected")
        self.driver.set_elm_unlock_pos()

    def Home(self, params: Command.Home) -> None:
        if not self.driver:
            raise Exception("Bioshake driver not connected")
        self.driver.home()

    def StartShake(self, params: Command.StartShake) -> None:

        if params.duration > 0:
            self.driver.shake_on_with_runtime(
                seconds=params.duration,
                speed=params.speed,
                acceleration=params.acceleration,
            )
        else:
            self.driver.start_shake(
                speed=params.speed,
                acceleration=params.acceleration,
            )

    def StopShake(self, params: Command.StopShake) -> None:
        self.driver.stop_shake()

    def WaitForShakeToFinish(self, params: Command.WaitForShakeToFinish) -> None:
        self.driver.wait_for_shake(timeout=params.timeout)

    def Reset(self, params: Command.Reset) -> None:
        if not self.driver:
            raise Exception("Bioshake driver not connected")
        self.driver.reset()

    def TemperatureOn(self, params: Command.TemperatureOn) -> None:
        self.driver.temp_on()

    def TemperatureOff(self, params: Command.TemperatureOff) -> None:
        self.driver.temp_off()

    def SetTemperature(self, params: Command.SetTemperature) -> None:
        self.driver.set_tmp(params.temperature)
    
    def EstimateGrip(self, params: Command.Grip) -> int:
        return 4

    def EstimateUngrip(self, params: Command.Ungrip) -> int:
        return 4

    def EstimateHome(self, params: Command.Home) -> int:
        return 10

    def EstimateStartShake(self, params: Command.StartShake) -> int:
        return params.duration

    def EstimateStopShake(self, params: Command.StopShake) -> int:
        return 10

    def EstimateReset(self, params: Command.Reset) -> int:
        return 2

    def EstimateWaitForShakeToFinish(self, params: Command.WaitForShakeToFinish) -> int:
        return 5


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
        raise RuntimeWarning("Port must be provided...")
    serve(BioShakeServer(), str(args.port))



#from tools.base_server import ABCToolDriver
import os
import time
import subprocess
import logging 
from typing import Optional 
from os.path import join
from tools.base_server import ABCToolDriver

RUN_CONTROL_SDK = "C:\\Program Files (x86)\\Hamilton\\Bin\\HxRun.exe"
DIR_NAME = os.path.dirname(os.path.abspath(__file__))
VENUS_LOG = join(DIR_NAME, "venus.log")


class HamiltonDriver(ABCToolDriver):

    def __init__(self)-> None:
        if not os.path.exists(RUN_CONTROL_SDK):
            raise ImportError(f"Failed to find {RUN_CONTROL_SDK}. Make sure that the path is correct.")
        self.protocol_name : Optional[str] = None
        self.live : bool  = False
        self.hamilton_process : Optional[subprocess.Popen] = None
        self.kill_processes("HsRun.exe")

    def remove_venus_log(self) -> None:
        try:
            if os.path.exists(VENUS_LOG):
                os.remove(VENUS_LOG)
        except Exception as ex:
            raise RuntimeError(ex)
        return None
    
    def load_and_run_protocol(self, protocol_name: str, close_on_end:bool =True)-> None:
        try:
            self.kill_processes("HxRun.exe",False)
            self.remove_venus_log()
            cmd  = [RUN_CONTROL_SDK, protocol_name, "-r"]
            if close_on_end:
                cmd.append("-t")
            self.hamilton_process = subprocess.Popen(cmd,stdout=open("venus.log",'a'), stderr=subprocess.STDOUT,  universal_newlines=True)
        except subprocess.CalledProcessError as ex:
            raise RuntimeError(f"Failed to run {protocol_name}'. Error Message: {str(ex)}")
        return None 
    
    def wait_for_protocol(self, timeout:int=500) -> None:
        start_time = time.time()
        logging.info(f"Run protocol started at {start_time}")
        seconds_spent_waiting = 0 
        while True:
            seconds_spent_waiting = int(time.time() - start_time)
            if self.hamilton_process is None:
                break
            status = self.hamilton_process.poll()
            if status is not None:
                break
            if timeout and seconds_spent_waiting > timeout:
                raise Exception("Run protocol has timed out. Please reset VWorks and restart the driver.")
            if os.path.exists(VENUS_LOG):
                try:
                    with open(join(DIR_NAME, "venus.log")) as file:
                        lines = file.readlines()
                        if len(lines) > 0:
                            raise RuntimeError(f"Encountered a run error {','.join(lines)}")
                except Exception as e:
                    raise RuntimeError(f"Encounter a runtime error. Error={e}")
            time.sleep(1)  # Add a small delay to avoid busy-waiting
        logging.info(f"Run protocol completed after {seconds_spent_waiting} seconds")
        return None 
    
    def load_protocol(self, protocol_name: str) -> None:
        if not os.path.exists(protocol_name):
            raise Exception(f"Protocol {protocol_name} not found.")
        self.kill_processes("HsRun.exe",False)
        self.remove_venus_log()
        cmd  = [RUN_CONTROL_SDK, protocol_name]
        try:
            self.hamilton_process = subprocess.Popen(cmd,stdout=open("venus.log",'a'), stderr=subprocess.STDOUT,  universal_newlines=True)
        except Exception as e:
            raise RuntimeError(f"Failed to load protocol {protocol_name}.Reason={e}")
        self.protocol_name = protocol_name

    def run_protocol(self, protocol_name:str) -> None:
        if not os.path.exists(protocol_name):
            raise Exception(f"Protocol {protocol_name} not found.")
        self.load_and_run_protocol(protocol_name)
        self.wait_for_protocol()
        return None



import logging
from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.microserve_pb2 import Command, Config
from .driver import MicroServeDriver

import argparse

class MicroserveServer(ToolServer):
    toolType = "microserve"
    driver: MicroServeDriver
    config: Config

    def __init__(self) -> None:
        super().__init__()

    def _configure(self, config: Config) -> None:
        self.config = config
        if self.driver:
            self.driver.disconnect()
        self.driver = MicroServeDriver(self.config.ip, self.config.port)
        self.driver.connect()
        
    def Load(self, params: Command.Load) -> None:
        self.driver.load(params.stack_id, params.plate_height, params.plate_thickness, params.stack_height)

    def Unload(self, params: Command.Unload) -> None:
        self.driver.unload(params.stack_id, params.plate_height, params.plate_thickness, params.stack_height)

    def Home(self, params: Command.Home) -> None:
        self.driver.home()

    def Retract(self, params:Command.Retract) -> None:
        self.driver.retract()

    def GoTo(self, params:Command.GoTo) -> None:
        self.driver.go_to(params.stack_id)

    def Abort(self, params:Command.Abort) -> None:
        self.driver.abort()

    def SendRawCommand(self, params: Command.SendRawCommand) -> None:
        logging.info(f"Sending raw command {params.command}")
        self.driver.send_command(params.command)

    def EstimatateLoad(self, params:Command.Load) -> int:
        return 1 
    
    def EstimateUnload(self, params:Command.Unload) -> int:
        return 1
    
    def EstimateHome(self, params:Command.Home) -> int:
        return 1
    
    def EstimateRetract(self, params:Command.Retract) -> int:
        return 1
    
    def EstimateGoTo(self, params:Command.GoTo) -> int:
        return 1
    
    def EstimateAbort(self, params:Command.Abort) -> int:
        return 1
    
    def EstimateSendRawCommand(self, params:Command.SendRawCommand) -> int:
        return 1
    
if __name__ == "__main__":
    #logging.basicConfig(level=logging.DEBUG)
    logging.getLogger().setLevel(logging.INFO)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    logging.info("Running server")
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(MicroserveServer(), str(args.port))



import logging

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.xpeel_pb2 import Command, Config  # Adjust the import paths according to your actual protobuf files

from tools.grpc_interfaces.tool_base_pb2 import ExecuteCommandReply
from tools.grpc_interfaces.tool_base_pb2 import INVALID_ARGUMENTS, SUCCESS

from .driver import XPeelDriver
import argparse 

class XPeelServer(ToolServer):
    toolType = "xpeel"
    driver: XPeelDriver
    config: Config

    def __init__(self) -> None:
        super().__init__()

    def _configure(self, config: Config) -> None:
        self.config = config
        if self.driver:
            self.driver.close()
        self.driver = XPeelDriver(self.config.com_port)

    def Peel(self, params: Command.Peel) -> None:
        response = ExecuteCommandReply()
        response.return_reply = True
        response.response = SUCCESS
        try:
            self.driver.remove_seal()
        except Exception as exc:
            logging.exception("Failed to deseal plate", exc)
            response.response = INVALID_ARGUMENTS

    def CheckStatus(self, params: Command.CheckStatus) -> ExecuteCommandReply:
        response = ExecuteCommandReply()
        response.return_reply = True
        response.response = SUCCESS
        try:
            self.driver.check_status()
        except Exception as exc:
            logging.exception("Failed to check status", exc)
            response.response = INVALID_ARGUMENTS
        return response

    def Reset(self, params: Command.Reset) -> ExecuteCommandReply:
        response = ExecuteCommandReply()
        response.return_reply = True
        response.response = SUCCESS
        try:
            self.driver.reset()
        except Exception as exc:
            logging.exception("Failed to reset device", exc)
            response.response = INVALID_ARGUMENTS
        return response

    def Restart(self, params: Command.Restart) -> ExecuteCommandReply:
        response = ExecuteCommandReply()
        response.return_reply = True
        response.response = SUCCESS
        try:
            self.driver.restart()
        except Exception as exc:
            logging.exception("Failed to restart device", exc)
            response.response = INVALID_ARGUMENTS
        return response

    def GetRemainingTape(self, params: Command.GetRemainingTape) -> ExecuteCommandReply:
        response = ExecuteCommandReply()
        response.return_reply = True
        response.response = SUCCESS
        try:
            self.driver.check_tape_remaining()
        except Exception as exc:
            logging.exception("Failed to check tape remaining", exc)
            response.response = INVALID_ARGUMENTS
        return response

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
        raise RuntimeWarning("Port must be provided...")
    serve(XPeelServer(), str(args.port))



"""Pydantic Schemas for waypoints.json"""

from pydantic.v1 import BaseModel # type: ignore
import typing as t

class Coordinate(str):
    @classmethod
    def __get_validators__(cls) -> t.Generator[t.Callable[..., "Coordinate"], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, v: t.Any) -> "Coordinate":
        if not isinstance(v, str):
            raise TypeError("Coordinate must be a string")
        if not v.count(" ") >= 4:
            raise ValueError("Coordinate must have at least 5 values separated by spaces")
        return Coordinate(v)

    def __repr__(self) -> str:
        return f"Coordinate({super().__repr__()})"

    @property
    def vec(self) -> list[float]:
        return [float(x) for x in self.split(" ")]

    def __add__(self, other: "str | Coordinate") -> "Coordinate":
        if not isinstance(other, Coordinate):
            other = Coordinate(other)
        return Coordinate(" ".join([str(a + b) for a, b in zip(self.vec, other.vec)]))

    def distance_to(self, other: "str | Coordinate") -> float:
        if not isinstance(other, Coordinate):
            other = Coordinate(other)
        distance: float = sum([(a - b) ** 2 for a, b in zip(self.vec, other.vec)]) ** 0.5
        return distance

    def __sub__(self, other: "str | Coordinate") -> "Coordinate":
        if not isinstance(other, Coordinate):
            other = Coordinate(other)
        return Coordinate(" ".join([str(a - b) for a, b in zip(self.vec, other.vec)]))


class Location(BaseModel):
    name: str
    tool_id: int
    id: int
    coordinates: Coordinate
    location_type: t.Literal["j", "c"]
    orientation: t.Literal["portrait", "landscape"]

class MotionProfile(BaseModel):
    id: int  # motion profile id
    name: str # motion profile name
    speed: float  # peak motion speed as percentage
    speed2: float  # secondary peak motion speed as percentage for cartesian moves
    acceleration: float  # peak motion acceleration as percentage
    deceleration: float  # peak motion deceleration as percentage
    accel_ramp: float  # acceleration ramp time in seconds
    decel_ramp: float  # deceleration ramp time in seconds
    inrange: float  # Distance from target to consider motion complete
    straight: int  # Boolean determining if the motion is curved/0 or straight/-1

    def __str__(self) -> str:
        return f"{self.id} {self.speed} {self.speed2} {self.acceleration} {self.deceleration} {self.accel_ramp} {self.decel_ramp} {self.inrange} {self.straight}"

class MotionProfiles(BaseModel):
    profiles: list[MotionProfile]

class Labware(BaseModel):
    id: int
    name: str
    description: str
    number_of_rows: int
    number_of_columns: int
    z_offset: float
    width : float
    height: float 
    plate_lid_offset: float #offset when the lid is on the plate
    lid_offset: float #offset when lid is on nest
    stack_height: float
    has_lid: bool

    image_url: t.Optional[str] = ""  # Allow image_url to be null or omitted

    class Config:
        # Allow extra fields like "created_at", "updated_at"
        extra = "allow"

class Labwares(BaseModel):
    labwares: list[Labware]

class Grip(BaseModel):
    id: int
    width: int
    speed: int 
    force: int 
    name: str
    tool_id: int

class Grips(BaseModel):
    grip_params: list[Grip]
    
class Waypoints(BaseModel):
    locations: list[Location]

class SequenceCommand(BaseModel):
    command: str
    params: t.Optional[dict[str, t.Any]]

class ArmSequence(BaseModel):
    name: str
    description: str
    commands: list[SequenceCommand]

class ArmSequences(BaseModel):
    sequences: list[ArmSequence]



import requests
from typing import Union, Any
import logging

API_URL = "http://localhost:8000/api"

class Db:
    @staticmethod
    def check_connection() -> bool:
        try:
            response = requests.get(API_URL)
            if response.status_code == 200:
                return True
        except requests.exceptions.ConnectionError:
            return False
        return False
    
    @staticmethod
    def get_data(model:str) -> Any:
        response = requests.get(f"{API_URL}/{model}")
        return response.json()

    @staticmethod
    def get_by_id_or_name(id:Union[int,str], model:str) -> Any:
        response = requests.get(f"{API_URL}/{model}/{id}")
        if response.status_code == 404:
            logging.warning(f"Resource with id/name {id} not found in {model}.")
            return None
        return response.json()
    
    @staticmethod
    def post_data(data:dict, model:str) -> Any:
        response = requests.post(f"{API_URL}/{model}", json=data)
        return response.json()

    @staticmethod
    def delete_data(id:Union[int,str], model:str) -> Any:
        response = requests.delete(f"{API_URL}/{model}/{id}")
        return response.json()
    
    @staticmethod
    def update_data(id:Union[str,int], data:dict, model:str) -> Any:
        response = requests.put(f"{API_URL}/{model}/{id}", json=data)
        return response.json()
    
    @staticmethod
    def ping(times:int) -> bool:
        for i in range(times):
            try:
                response = requests.get(f"{API_URL}/health")
                logging.info(f"Response status code: {response.status_code}")
                if response.status_code == 200:
                    return True
            except requests.exceptions.ConnectionError:
                continue
        logging.error("Could not establish connection to database...")
        return False
    


from tools.toolbox.db import Db 
from typing import Any

db = Db()

def get_workcell(id:int) -> Any:
    response = db.get_by_id_or_name(id, "workcells")
    return response
    
def get_all_workcells() -> Any:
    response = db.get_data("workcells")
    return response




import logging

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.hamilton_pb2 import Command, Config
from .driver import HamiltonDriver 

import argparse

class HamiltonServer(ToolServer):
    toolType = "hamilton"

    def __init__(self) -> None:
        super().__init__()
        self.driver : HamiltonDriver

    def _configure(self, request:Config) -> None:
        logging.info("Initializing Hamilton STAR")
        self.driver = HamiltonDriver()

    def RunProtocol(self, params:Command.RunProtocol) -> None:
        self.driver.run_protocol(params.protocol)

    def LoadProtocol(self, params:Command.LoadProtocol) -> None:
        self.driver.load_protocol(params.protocol)

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(HamiltonServer(), str(args.port))



"""
Galago Tools version information.
This file is auto-updated by the bump script.
"""

# Version information
__version__ = '0.13.0'  # Update this version string as needed

# Parse version components for easier access
VERSION_PARTS = __version__.split('.')
VERSION_MAJOR = int(VERSION_PARTS[0])
VERSION_MINOR = int(VERSION_PARTS[1])
VERSION_PATCH = int(VERSION_PARTS[2]) if len(VERSION_PARTS) > 2 else 0

# String representations
VERSION_SHORT = f"{VERSION_MAJOR}.{VERSION_MINOR}"
VERSION_FULL = __version__


import logging
import time

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.liconic_pb2 import Command, Config

from tools.grpc_interfaces.tool_base_pb2 import ExecuteCommandReply
from tools.grpc_interfaces.tool_base_pb2 import SUCCESS

from .driver import LiconicStxDriver

import argparse

class LiconicServer(ToolServer):
    toolType = "liconic"
    driver: LiconicStxDriver
    config: Config

    def __init__(self) -> None:
        super().__init__()

    def _configure(self, config: Config) -> None:
        self.config = config
        if self.driver:
            self.driver.close()
        self.driver = LiconicStxDriver(self.config.com_port)
        self.driver.start_monitor()
        
    def FetchPlate(self, params: Command.FetchPlate) -> None:
        response = ExecuteCommandReply()
        response.return_reply = True
        response.response = SUCCESS
        #try:            
        if params.cassette < 1 or params.level < 1:
            raise ValueError(
                f"Cassette ({params.cassette}) and level ({params.level}) must be positive integers"
            )
        self.driver.unload_plate(params.cassette, params.level)
        time.sleep(6)

    def StorePlate(self, params: Command.StorePlate) -> None:
        response = ExecuteCommandReply()
        response.return_reply = True
        response.response = SUCCESS
        #try:            
        if params.cassette < 1 or params.level < 1:
            raise ValueError(
                f"Cassette ({params.cassette}) and level ({params.level}) must be positive integers"
            )
        self.driver.load_plate(params.cassette, params.level)
        time.sleep(6)


    def Reset(self, params: Command.Reset) -> None:
        self.driver.reset()

    def EstimateFetchPlate(self, params: Command.FetchPlate) -> int:
        return 1

    def EstimateStorePlate(self, params: Command.StorePlate) -> int:
        return 1

    def EstimateReset(self, params: Command.Reset) -> int:
        return 1

    def SendRawCommand(self, params: Command.SendRawCommand) -> None:
        logging.info(f"Sending raw command {params.cmd}")
        return self.driver.raw(params.cmd)

if __name__ == "__main__":
    logging.getLogger().setLevel(logging.INFO)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    logging.info("Running server")
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(LiconicServer(), str(args.port))


from datetime import datetime
import os
import logging
from enum import Enum 
from typing import Optional, Any
import sys 
import socket 

class LogType(Enum):
    ERROR = "ERROR",
    WARNING = "WARNING",
    DEBUG = "DEBUG",
    INFO = "INFO",
    PLATE_MOVE = "PLATE_MOVE",
    RUN_START = "RUN_START",
    RUN_END = "RUN_END",
    PLATE_READ = "PLATE_READ",

def get_local_ip() -> Any:
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        local_ip = s.getsockname()[0]
    except Exception:
        local_ip = "127.0.0.1" # Default local IP
    finally:
        s.close()
    return local_ip

def get_shell_command(tool_type:str, port:int) -> list:
    return [sys.executable, '-m', f'tools.{tool_type}.server', f'--port={port}']

def write_trace_log(log_path:Optional[str], log_type:LogType, tool:str,value:str) -> None:

    if not log_path:
        logging.warning("Log folder not configured")
    if log_path is None:
        return
    if not os.path.exists(log_path):
        try:
            os.makedirs(log_path)
        except Exception as e:
            logging.warning(f"Failed to create log folder. {e}")
            return
    file_folder= os.path.join(log_path, datetime.today().strftime('%Y-%m-%d'))
    if(os.path.exists(file_folder) is False):
        logging.debug("folder does not exist. creating folder")
        os.mkdir(file_folder)

    trace_file = os.path.join(file_folder, "trace_log.txt")
    error_file = os.path.join(file_folder, "error_log.txt")

    try:
        if os.path.exists(trace_file) is False:
            with open(trace_file, 'w+') as f:
                f.write('Time,Tool,Value\n')
        if(log_type == LogType.ERROR):
            if os.path.exists(error_file) is False:
                with open(error_file, 'w+') as f:
                    f.write('Time,Tool,Error\n')
    except Exception as e:
        logging.debug(e)
        return
    try:
        with open(trace_file, 'a') as f:
            f.write(str(datetime.today())+","+str(log_type.name)+","+tool+","+value+"\n")
        if log_type == LogType.ERROR:
            with open(error_file, 'a') as f:
                 f.write(str(datetime.today())+","+str(log_type.name)+","+tool+","+value+"\n")
    except Exception as e:
        logging.debug(e)
        return




import logging.handlers
import logging
import os
import sys
import subprocess
from tools.app_config import Config
import socket 
import signal as os_signal
import time
import argparse
from os.path import dirname
from tools.utils import get_shell_command 

ROOT_DIR = dirname(dirname(os.path.realpath(__file__)))
LOG_TIME = int(time.time())
TOOLS_32BITS = ["vcode","bravo","hig_centrifuge","plateloc","vspin"]

sys.path = [
    p for p in sys.path
    if not any(sub in p.lower() for sub in ["anaconda3", "miniconda", "mamba"])
]

logging.basicConfig(
    level=logging.DEBUG,
    format='%(asctime)s | %(levelname)s | %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S', 
)

class LaunchConsole():
    def __init__(self, config:Config) -> None:
        self.running_tools = 0
        self.config_file = ""
        logging.info("Starting Galago Manager")
        self.config :Config = config
        working_dir = "" 
        self.log_folder = os.path.join(working_dir,"data","trace_logs", str(LOG_TIME))
        self.server_processes : dict[str,subprocess.Popen] = {}
        
        if not os.path.exists(self.log_folder):
            os.makedirs(self.log_folder)

    def kill_all_processes(self) ->None:
        logging.info("Killing all processes")
        for proc_key, process in self.server_processes.items():
            try:
                self.kill_by_process_id(process.pid)
                time.sleep(0.5)
                logging.info(f"Killed process {process.pid}")
                del process
            except ProcessLookupError as e:
                logging.error(f"failed to shut down process. Error={str(e)}")
                pass
        self.server_processes.clear()
        self.force_kill_tool()
    
    def load_tools(self) -> None:
        self.config.load_workcell_config()
 
    def read_last_lines(self, filename:str, lines:int=100) -> list[str]:
        with open(filename, 'rb') as f:
            f.seek(0, os.SEEK_END)
            end_position = f.tell()
            buffer_size = 1024
            blocks = -1
            data = []
            while end_position > 0 and len(data) < lines:
                if end_position - buffer_size > 0:
                    f.seek(blocks * buffer_size, os.SEEK_END)
                else:
                    f.seek(0, os.SEEK_SET)
                data.extend(f.readlines())
                end_position -= buffer_size
                blocks -= 1
            return [line.decode('utf-8') for line in data[-lines:]]
    
    def __del__(self) -> None:
        self.kill_all_processes()
    
    def kill_by_process_id(self, process_id:int) -> None:
        try:
            if os.name == 'nt':
                subprocess.call(['taskkill', '/F', '/T', '/PID', str(process_id)])
            else:
                os.kill(process_id, os_signal.SIGINT)
        except ChildProcessError as e:
            logging.info(f"Failed to kill process {process_id}. Reason is={str(e)}")
        finally:
            return None

    def run_subprocess(self, tool_type:str, tool_name:str, port:int,confirm_modal:bool=False) -> None:
        try:
            self.kill_process_by_name(str(tool_name))
            tool_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = tool_socket.connect_ex(('127.0.0.1',int(port)))
            if result != 0:
                cmd = get_shell_command(tool_type=tool_type, port=port)
                os.chdir(ROOT_DIR)
                use_shell = False
                if os.name == 'nt':
                    use_shell = True
                logging.info(f"log folder is {self.log_folder}")
                # if self.log_folder:
                #     output_file = join(self.log_folder, str(tool_name)) + ".log"
                #     process = subprocess.Popen(cmd, stdout=open(output_file,'w'), stderr=subprocess.STDOUT,  universal_newlines=True)
                # else:
                process = subprocess.Popen(cmd, shell=use_shell,universal_newlines=True)
                self.server_processes[tool_name] = process
            else:
                logging.warning(f"Port {port} for {tool_name} is already occupied")
        except subprocess.CalledProcessError:
            logging.info("There was an error launching tool server.")
        return None
    
    def kill_process_by_name(self, process_name:str) -> None:
        if process_name not in self.server_processes.keys():
            return None
        else:
            try:
                process_id = self.server_processes[process_name].pid
                self.kill_by_process_id(process_id)
            except Exception as e:
                logging.warning(f"Failed to kill process {process_name}. Reason is={str(e)}.")
        return None 
    
    def force_kill_tool(self) -> None:
        try:
            if os.name != 'nt':
                subprocess.Popen("lsof -t -i tcp:1010 | xargs kill", shell=True)
        except Exception as e:
            logging.error(f"Failed to force kill toolbox. Reason is {str(e)}")

    def start_toolbox(self) -> None:
        logging.info("Launching Toolbox")
        try:
            self.run_subprocess("toolbox", "Tool Box",1010,False)
        except subprocess.CalledProcessError:
            logging.info("There was an error launching toolbox server.")

    def run_all_tools(self) -> None:
        self.kill_all_processes()
        time.sleep(0.5)
        self.load_tools()
        self.start_toolbox()
        if self.config.workcell_config is None:
            logging.error("No workcell configuration loaded")
            return

        for t in self.config.workcell_config.tools:
            logging.info(f"Launching process for tool {t.name}")
            #Check if tool is already running. 
            tool_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = tool_socket.connect_ex(('127.0.0.1',t.port))
            if result != 0:
                try:
                    self.run_subprocess(t.type,t.name,t.port,False )
                except Exception as e:
                    logging.error(f"Failed to launch tool {t.name}. Error is {e}")
            else:
                logging.warning(f"Port for tool {t.name} is already occupied")
        time.sleep(0.5)

def main() -> int:
    parser = argparse.ArgumentParser(description='Launch Galago Tools Manager')
    parser.add_argument('--debug', action='store_true', help='Enable debug mode with detailed logging')
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(
            level=logging.DEBUG,
            format='%(asctime)s | %(levelname)s | %(message)s',
            handlers=[
                logging.StreamHandler(sys.stdout)
            ]
        )
    else:
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s | %(levelname)s | %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
        )

    try:
        config = Config()
        #config.load_app_config()
        logging.info("Loading workcell config")
        config.load_workcell_config()
        manager = LaunchConsole(config)
        manager.run_all_tools()

        logging.info("Tool servers are running. Press Ctrl+C to exit.")
        while True:
            time.sleep(1)
        return 0
    except KeyboardInterrupt:
        logging.info("Shutdown requested by user.")
        manager.kill_all_processes()
        return 0
    except Exception:
        logging.exception("Failed to launch tools")
        sys.exit(1)
        return 1
    
if __name__ == "__main__":
    main()


from tools.toolbox.db import Db
from typing import Any, Dict

db = Db()

def get_all_logs() -> Any:
    """Get all logs from the database."""
    response = db.get_data("logs")
    return response

def get_paginated_logs(skip: int, limit: int, descending: bool) -> Any:
    """Get paginated logs from the database.
    
    Args:
        skip: Number of logs to skip
        limit: Maximum number of logs to return
        descending: Whether to sort in descending order
    """
    response = db.get_data(f"logs?skip={skip}&limit={limit}&descending={descending}")
    return response

def clear_all_logs() -> Any:
    """Clear all logs from the database."""
    # The delete_data method typically requires an ID, but in this case,
    # we're deleting all logs, so we'll use a special endpoint
    response = db.delete_data("", "logs")
    return response

def add_log(log_data: Dict[str, Any]) -> Any:
    """Add a new log to the database.
    
    Args:
        log_data: Dictionary containing log information
    """
    response = db.post_data(log_data, "logs")
    return response



import json 
from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.pf400_pb2 import Command, Config
from .driver import Pf400Driver
import argparse
from typing import Optional, Union 
from tools.grpc_interfaces.tool_base_pb2 import ExecuteCommandReply, SUCCESS, ERROR_FROM_TOOL
from google.protobuf.struct_pb2 import Struct
import logging
from tools.pf400.waypoints_models import (
    Waypoints,
    MotionProfiles,
    MotionProfile,
    Grips,
    Grip,
    Location,
    Labwares,
    ArmSequences,
    ArmSequence,
    Coordinate,
    Labware
)       
from google.protobuf.json_format import MessageToDict
from google.protobuf import json_format
from google.protobuf import message
import typing as t  
from google.protobuf.json_format import Parse

#Default motion profiles
DEFAULT_MOTION_PROFILES : list[MotionProfile] = [
    #curved motion profile
    MotionProfile(
        name="default",
        id=1,
        speed=85,
        speed2=80,
        acceleration=60,
        deceleration=60,
        accel_ramp=0.1,
        decel_ramp=0.1,
        inrange=0,
        straight=0
    ), 
    #straight motion profile
    MotionProfile(
        name="default_straight",
        id=2,
        speed=85,
        speed2=80,
        acceleration=100,
        deceleration=100,
        accel_ramp=0.1,
        decel_ramp=0.1,
        inrange=0,
        straight=1
    ),
]


DEFAULT_PLATE_HANDLING_PARAMS : dict[str, dict[str, Union[Command.GraspPlate, Command.ReleasePlate]]]= {
    "landscape":
    {
        "grasp": Command.GraspPlate(width=122, force=15, speed=10),
        "release": Command.ReleasePlate(width=130, speed=10)
    },
    "portrait": {
        "grasp": Command.GraspPlate(width=86, force=15, speed=10),
        "release": Command.ReleasePlate(width=96, speed=10)
    }
}

class Pf400Server(ToolServer):
    toolType = "pf400"

    def __init__(self) -> None:
        super().__init__()
        self.config : Config 
        self.driver: Pf400Driver
        self.waypoints: Waypoints
        self.labwares : Labwares
        self.sequences : ArmSequences
        self.motion_profiles : MotionProfiles
        self.grips : Grips
        self.plate_handling_params : dict[str, dict[str, Union[Command.GraspPlate, Command.ReleasePlate]]] = {}

    def _configure(self, request: Config) -> None:
        self.config = request
        if self.driver:
            self.driver.close()
        self.driver = Pf400Driver(
            tcp_host=request.host,
            tcp_port=request.port,
            joints=request.joints,
            gpl_version=request.gpl_version,

        )
        self.driver.initialize()

    def _getGrip(self, grip_name:str) -> Grip:
        grip = next((x for x in self.grips.grip_params if x.name.lower() == grip_name.lower()), None)
        if not grip:
            raise RuntimeError(f"Grip {grip_name} not defined.")
        return grip

    def _getLocation(self, location_name: str) -> Optional[Location]:
        location = next((x for x in self.waypoints.locations if x.name.lower() == location_name.lower()), None)
        if not location:
            logging.warning(f"Location '{location_name.lower()}' not found")
        return location
    
    def _getLabware(self, labware_name: str) -> Labware:
        labware = next((x for x in self.labwares.labwares if x.name == labware_name), None)
        if not labware:
            raise Exception(f"Labware '{labware_name}' not found")
        return labware
    
    def _getSequence(self, sequence_name:str) -> ArmSequence: 
        sequence = next((x for x in self.sequences.sequences if x.name == sequence_name), None)
        if not sequence:
            raise Exception(f"Sequence '{sequence_name}' not found")
        logging.info(f"Found sequence! {sequence_name}")
        return sequence
    
    def _getProfileId(self, profile_name:str) -> int:
        try:
            logging.info(f"Searching for profile {profile_name}")
            profile = next((x for x in self.motion_profiles.profiles if x.name.lower() == profile_name.lower()), None)
            if not profile:
                error_message = f"Profile '{profile_name}' not found"
                logging.error(error_message)
                return 1 # Default to the first profile if not found
            return profile.id
        except Exception as e:
            logging.error(f"Error in _getProfileId: {str(e)}")
            raise


    def LoadWaypoints(self, params: Command.LoadWaypoints) -> None:
            logging.info("Loading waypoints")
            #Load locations
            waypoints_dictionary: dict[str, t.Any] = json_format.MessageToDict(params.waypoints)
            locations_list = waypoints_dictionary.get("locations", [])
            self.waypoints = Waypoints.parse_obj({"locations": locations_list})
            logging.info(f"Loaded {len(self.waypoints.locations)} locations")

            #Load grips
            grips_list = waypoints_dictionary.get("grip_params")
            grip_params : Grips = Grips.parse_obj({"grip_params": grips_list})
            logging.info(f"Loaded {len(grip_params.grip_params)} grip parameters")
            if len(grip_params.grip_params) == 0:
                logging.warning("No grip parameters found. Using default grip parameters.")
                self.plate_handling_params = DEFAULT_PLATE_HANDLING_PARAMS
            else:
                for grip in grip_params.grip_params:
                    self.plate_handling_params[grip.name.lower()] = {
                        "grasp": Command.GraspPlate(width=grip.width, force=grip.force, speed=grip.speed),
                        "release": Command.ReleasePlate(width=grip.width+10, speed=grip.speed)
                    }

            self.grips = grip_params

            logging.info(f"Loaded {len(self.plate_handling_params)} plate handling parameters") 
            logging.info(self.plate_handling_params)
            
            #Load and register profiles 
            motion_profiles_list = waypoints_dictionary.get("motion_profiles")
            if motion_profiles_list:
                for i, profile in enumerate(motion_profiles_list):
                    profile["id"] = i + 1
            motion_profiles = MotionProfiles.parse_obj({"profiles": motion_profiles_list})
            
            self.motion_profiles = motion_profiles
            logging.info(f"Loaded {len(motion_profiles.profiles)} motion profiles")
  
            if motion_profiles_list and len(motion_profiles_list) > 0:
                for motion_profile in motion_profiles.profiles:
                    logging.info(f"Registering motion profile {motion_profile.name}")
                    try:
                        profile_no_name = motion_profile.copy(deep=True)
                        profile_no_name.name = ""
                        self.driver.register_motion_profile(str(motion_profile))
                    except Exception as e:
                        logging.error(f"Error registering motion profile {motion_profile.name}: {e}")
                        raise Exception(f"Error registering motion profile {motion_profile.name}: {e}")
            else:
                #Register default motion profiles
                logging.info("No motion profiles loaded. Using default profiles.")
                for motion_profile in DEFAULT_MOTION_PROFILES:
                    logging.info(f"Registering default motion profiles {motion_profile.name}")
                    try:
                        #Remove name from the profile 
                        profile_no_name = motion_profile.copy(deep=True)
                        profile_no_name.name = ""
                        logging.info(profile_no_name)
                        self.driver.register_motion_profile(str(motion_profile))
                    except Exception as e:
                        logging.error(f"Error registering motion profile {motion_profile.name}: {e}")
                        raise Exception(f"Error registering motion profile {motion_profile.name}: {e}")
                
            # #Load Sequences 
            sequences_list = waypoints_dictionary.get("sequences")
            sequences = ArmSequences.parse_obj({"sequences":sequences_list})
            logging.info(f"Loaded {len(sequences.sequences)} sequences")
            self.sequences = sequences

    
    def LoadLabware(self, params: Command.LoadLabware) -> None:
        logging.info("Loading labware")
        labware_dictionary = MessageToDict(params.labwares)
        labware_lst = labware_dictionary.get("labwares")
        self.labwares = Labwares.parse_obj({"labwares": labware_lst})
        logging.info(f"Loaded {len(self.labwares.labwares)} labwares")

    def _unwind(self) -> None:
        location = next((x for x in self.waypoints.locations if x.name.lower() in {"unwind"}), None)

        if not location:
            raise KeyError("Unwind location not found. Please add a unwind location to the waypoints. Height or rail position does not matter.")
        
        waypoint_loc = location.coordinates
        current_loc_array = self.driver.wherej().split(" ")
        #Unwind the arm while keeping the z height, gripper width and rail constant
        if self.config.joints == 5:
           new_loc = f"{current_loc_array[1]} {waypoint_loc.vec[1]} {waypoint_loc.vec[2]} {waypoint_loc.vec[3]} {current_loc_array[5]}"
        else:
            new_loc = f"{current_loc_array[1]} {waypoint_loc.vec[1]} {waypoint_loc.vec[2]} {waypoint_loc.vec[3]} {current_loc_array[5]} {current_loc_array[6]}"
        self.driver.movej(new_loc,  motion_profile=1)

    def Unwind(self,params: Command.Unwind) -> None:
        self._unwind()

    def Release(self, params: Command.Release) -> None:
        self.driver.safe_free()

    def Engage(self, params: Command.Engage) -> None:
        self.driver.unfree()

    def moveTo(
        self,
        loc: Location,
        approach_height: float = 0, #Height to move to before moving to the specified location
        motion_profile_id: int = 1,
    ) -> None:
        if self.driver is None:
            return
        loc_type = loc.location_type
        if loc_type == "c":
            string_offset =  f"0 0 {approach_height} 0 0 0"
            if self.config.joints == 5:
                string_offset = " ".join(string_offset.split(" ")[:-1])
            self.driver.movec(
                str(loc.coordinates + Coordinate(string_offset)),
                motion_profile=motion_profile_id,
            )
        #For now we only handle a z offset for joints
        elif loc_type == "j":
            string_offset = f"{approach_height} 0 0 0 0 0"
            self.driver.movej(str(loc.coordinates + Coordinate(string_offset)), 
                                motion_profile=motion_profile_id)
        else:
            raise Exception("Invalid location type")
        
    def Move(self, params: Command.Move) -> None:
        """Execute a move command with the given coordinate and motion profile."""
        location = self._getLocation(params.location)
        profile_id = self._getProfileId(params.motion_profile)
        if location is None:
            raise Exception(f"Location '{params.location}' not found")
        self.moveTo(location, params.approach_height, motion_profile_id=profile_id)

    def GraspPlate(self, params: Command.GraspPlate) -> None:
        self.driver.graspplate(params.width, params.force, params.speed)

    def ReleasePlate(self, params: Command.ReleasePlate) -> None:
        self.driver.releaseplate(params.width, params.speed)

    def retrieve_plate(
        self,
        source_nest: str,
        grasp_params: Optional[Command.GraspPlate] = None,
        approach_height: float = 0,
        motion_profile: str = "default",
        grip_width: int = 0,
        labware_name: str = "",
    ) -> None:
        source_location = self._getLocation(source_nest)
        if not source_location:
            raise Exception(f"Location '{source_nest}' not found")

        labware = self._getLabware(f"{labware_name}")
        labware_offset = labware.z_offset
        labware_offset = int(labware_offset)
        grip_width = int(grip_width)
        approach_height = int(approach_height)
        grasp: Command.GraspPlate
        if not grasp_params or (grasp_params.width == 0):
            grap_param_exists = self.plate_handling_params.get(source_location.orientation.lower())
            if not grap_param_exists:
                raise Exception(f"Grasp params for {source_location.orientation.lower()} not found")
            tmp_grasp: Union[Command.GraspPlate,Command.ReleasePlate] = self.plate_handling_params[
                source_location.orientation.lower()
            ]["grasp"]
            if isinstance(tmp_grasp, Command.GraspPlate):
                grasp = tmp_grasp
            else:
                raise Exception("Invalid grasp params")
        else:
            grasp = grasp_params
        if grip_width > 0:
            adjust_gripper = Command.ReleasePlate(width=grip_width, speed=10)
        else:
            tmp_adjust_gripper = self.plate_handling_params[
                source_location.orientation.lower()
            ]["release"]
            if isinstance(tmp_adjust_gripper, Command.ReleasePlate):
                adjust_gripper = tmp_adjust_gripper
            else:
                raise Exception("Invalid release params")
            
        pre_grip_sequence : t.List[message.Message] = []
        retrieve_sequence : t.List[message.Message] = []
        open_grip_width= self._getGrip(source_location.orientation).width + 10
        pre_grip_sequence.append(adjust_gripper)
        pre_grip_sequence.append(Command.Move(location=source_location.name, motion_profile=motion_profile, approach_height=approach_height))
        pre_grip_sequence.append(Command.Move(location=source_location.name, motion_profile=motion_profile, approach_height=labware_offset))
        retrieve_sequence.extend([
            grasp,  # Grasp the plate
            Command.Move(location=source_location.name, motion_profile=motion_profile, approach_height=approach_height)
        ])
        self.driver.state.gripper_axis_override_value = open_grip_width
        self.runSequence(pre_grip_sequence)
        self.driver.state.gripper_axis_override_value = None
        self.runSequence(retrieve_sequence)

    def dropoff_plate(
        self,
        destination_nest: str,
        release_params: Optional[Command.ReleasePlate] = None,
        approach_height: float = 0,
        motion_profile: str = "default",
        labware_name: str = "",
    ) -> None:
        dest_location = self._getLocation(destination_nest)
        if not dest_location:
            raise Exception(f"Location '{destination_nest}' not found")

        release: Command.ReleasePlate
        labware = self._getLabware(f"{labware_name}")
        labware_offset = int(labware.z_offset)

        if not release_params or (release_params.width == 0):
            release_param_exists = self.plate_handling_params.get(dest_location.orientation.lower())
            if not release_param_exists:
                raise Exception(f"Release params for {dest_location.orientation.lower()} not found")
            tmp_release: Union[Command.GraspPlate , Command.ReleasePlate] = self.plate_handling_params[
                dest_location.orientation.lower()
            ]["release"]
            if isinstance(tmp_release, Command.ReleasePlate):
                release = tmp_release
            else:
                raise Exception("Invalid release params")
        else:
            release = release_params

        dropoff_sequence :t.List[message.Message] = []
        post_dropoff_sequence: t.List[message.Message] = []
        
        dropoff_sequence.extend([
                Command.Move(location=dest_location.name, motion_profile=motion_profile, approach_height=int(approach_height)), #Move to the dest location plus offset
                Command.Move(location=dest_location.name, motion_profile=motion_profile,approach_height=labware_offset), #Move to the nest location down in a straight pattern
                release #release the plate
        ])

        post_dropoff_sequence.append(Command.Move(location=dest_location.name, motion_profile=motion_profile, approach_height=int(approach_height))) #Move to the approach offset

        self.runSequence(dropoff_sequence)
        self.driver.state.gripper_axis_override_value = self._getGrip(dest_location.orientation).width + 10
        self.runSequence(post_dropoff_sequence)
        self.driver.state.gripper_axis_override_value = None

    def RetrievePlate(self, params: Command.RetrievePlate) -> None:
        self.retrieve_plate(source_nest=params.location, motion_profile=params.motion_profile, approach_height=params.approach_height, labware_name=params.labware)

    def DropOffPlate(self, params: Command.DropOffPlate) -> None:
        self.dropoff_plate(destination_nest=params.location, motion_profile=params.motion_profile, approach_height=params.approach_height, labware_name=params.labware)

    def Jog(self, params: Command.Jog) -> None:
        """Handle jog command from UI"""
        if not self.driver:
            raise Exception("Driver not initialized")
        self.driver.jog(params.axis, params.distance)

    def Transfer(self, params: Command.Transfer) -> None:
        self.retrieve_plate(
            source_nest=params.source_nest,
            motion_profile = params.motion_profile,
            approach_height=5,
            labware_name=params.labware
        )
        self._unwind()
        self.dropoff_plate(
            destination_nest=params.destination_nest,
            motion_profile=params.motion_profile,
            approach_height=5,
            labware_name=params.labware
        )


    def _pick_lid(
        self,
        location_name: str,
        labware_name: str,
        pick_from_plate: bool = False,
        approach_height: float = 0,
        motion_profile: str = "default",
    ) -> None:
        location: Optional[Location] = self._getLocation(location_name)
        if not location:
            raise Exception(f"Location '{location_name}' not found")
        
        # Get labware
        labware: Labware = self._getLabware(labware_name)
        
        # Configure grip parameters
        grasp: Command.GraspPlate
        tmp_grasp: Union[Command.GraspPlate, Command.ReleasePlate] = self.plate_handling_params[
            location.orientation.lower()
        ]["grasp"]
        if isinstance(tmp_grasp, Command.GraspPlate):
            grasp = tmp_grasp
            grasp.force = 15
            grasp.speed = 10
        else:
            raise Exception("Invalid grasp params")
        
        # Configure release parameters
        tmp_adjust_gripper = self.plate_handling_params[location.orientation.lower()]["release"]
        if isinstance(tmp_adjust_gripper, Command.ReleasePlate):
            adjust_gripper = tmp_adjust_gripper
        else:
            raise Exception("Invalid release params")
        
        # Calculate lid height
        if pick_from_plate:
            lid_height = labware.height - 6 + labware.plate_lid_offset
        else:
            lid_height = labware.plate_lid_offset + labware.lid_offset
        
        # Define sequences
        pre_pick_sequence: t.List[message.Message] = []
        pick_sequence: t.List[message.Message] = []
        
        # Configure gripper width
        open_grip_width = self._getGrip(location.orientation).width + 10
        
        # Set gripper override
        self.driver.state.gripper_axis_override_value = open_grip_width
        
        pre_pick_sequence.append(adjust_gripper)
        pre_pick_sequence.append(Command.Move(location=location.name, motion_profile=motion_profile, 
                                            approach_height=int(labware.height + approach_height)))
        pre_pick_sequence.append(Command.Move(location=location.name, motion_profile=motion_profile, 
                                            approach_height=int(lid_height)))
        
        # Build pick sequence
        pick_sequence.extend([
            grasp,
            Command.Move(location=location.name, motion_profile=motion_profile, approach_height=int(labware.height + approach_height)),
        ])
        
        
        # Execute sequences
        self.runSequence(pre_pick_sequence)
        self.driver.state.gripper_axis_override_value = None
        self.runSequence(pick_sequence)


    def PickLid(self, params: Command.PickLid) -> None:
        self._pick_lid(
            location_name=params.location,
            labware_name=params.labware,
            pick_from_plate=params.pick_from_plate,
            approach_height=params.approach_height,
            motion_profile=params.motion_profile
        )
    

    def _place_lid(
        self,
        location_name: str,
        labware_name: str,
        place_on_plate: bool = False,
        approach_height: float = 0,
        motion_profile: str = "default",
    ) -> None:
        # Get location
        location: Optional[Location] = self._getLocation(location_name)
        if not location:
            raise Exception(f"Location '{location_name}' not found")
        
        # Get labware
        labware: Labware = self._getLabware(labware_name)
        
        # Configure release parameters
        release: Command.ReleasePlate
        tmp_release: Union[Command.GraspPlate, Command.ReleasePlate] = self.plate_handling_params[
            location.orientation.lower()
        ]["release"]
        if isinstance(tmp_release, Command.ReleasePlate):
            release = tmp_release
        else:
            raise Exception("Invalid release params")
        
        # Calculate lid height
        if place_on_plate:
            lid_height = labware.height - 6 + labware.plate_lid_offset
        else:
            lid_height = labware.plate_lid_offset + labware.lid_offset
        
        # Define sequences
        place_lid_sequence: t.List[message.Message] = []
        post_place_sequence: t.List[message.Message] = []

        place_lid_sequence.extend([
            Command.Move(location=location.name, motion_profile=motion_profile, 
                        approach_height=int(labware.height + approach_height)),  # Move to the location plus approach height
            Command.Move(location=location.name, motion_profile=motion_profile, 
                        approach_height=int(lid_height)),  # Move to the calculated height
            release,
        ])
        
        post_place_sequence.append(Command.Move(location=location.name, motion_profile=motion_profile, 
                                            approach_height=int(labware.height + approach_height)))

        self.runSequence(place_lid_sequence)
        open_grip_width = self._getGrip(location.orientation).width + 10
        self.driver.state.gripper_axis_override_value = open_grip_width
        self.runSequence(post_place_sequence)
        
        self.driver.state.gripper_axis_override_value = None
        
    def PlaceLid(self, params: Command.PlaceLid) -> None:
        """Place lid handler that delegates to the _place_lid implementation"""
        self._place_lid(
            location_name=params.location,
            labware_name=params.labware,
            place_on_plate=params.place_on_plate,
            approach_height=params.approach_height,
            motion_profile=params.motion_profile
        )

    def GetCurrentLocation(self, params: Command.GetCurrentLocation) -> ExecuteCommandReply:
        """Get current robot position"""
        response = ExecuteCommandReply()
        response.return_reply = True
        response.response = SUCCESS
        try:
            if not self.driver:
                raise Exception("Driver not initialized")
            position = self.driver.wherej()  # Using the direct wherej command from driver
            
            # Create meta_data with the location information
            meta = Struct()
            meta.update({"location": position})
            response.meta_data.CopyFrom(meta)
            
            return response
        except Exception as e:
            logging.error(f"Error getting position: {e}")
            response.response = ERROR_FROM_TOOL
            response.error_message = str(e)
            return response

    def command_instance_from_name(self, command_name: str) -> Union[message.Message, t.Any]:
        command_descriptors = Command.DESCRIPTOR.fields_by_name
        command_dictionary = dict()
        for (field_name, field_descriptor) in command_descriptors.items():
            if field_descriptor.message_type:
                if field_name not in command_dictionary:
                    command_dictionary[field_name] = field_descriptor.message_type.name
        
        return command_dictionary[command_name]

    def RunSequence(self, params: Command.RunSequence) -> None:
        commandSequence : list[message.Message] = list()
        sequence = self._getSequence(params.sequence_name)
        logging.info(f"Sequence has {len(sequence.commands)} commands")
        for arm_command in sequence.commands:
            command_params :t.Any = arm_command.params

            command_name :str = arm_command.command
            command_type : t.Any = self.command_instance_from_name(command_name)
            command : t.Any  = getattr(Command, command_type)

            if command_name == 'dropoff_plate' or command_name == 'retrieve_plate'  or command_name == 'pick_lid' or command_name == 'place_lid':
                if params.labware is not None:
                    command_params["labware"] = params.labware
            try:
                command_parsed = Parse(json.dumps(command_params), command())
                commandSequence.append(command_parsed)
            except Exception as e:
                logging.error(f"Error parsing command {command_name} with params {command_params}: {e}")
                raise Exception(f"Error parsing command {command_name} with params {command_params}: {e}")
        self.runSequence(commandSequence)

    def estimateRelease(self, params: Command.Release) -> int:
        return 1

    def estimateEngage(self, params: Command.Engage) -> int:
        return 1

    def estimateUnwind(self, params: Command.Unwind) -> int:
        return 1
    
    def estimateGraspPlate(self, params: Command.GraspPlate) -> int:
        return 1

    def estimateReleasePlate(self, params: Command.ReleasePlate) -> int:
        return 1

    def estimateRunSequence(self, sequence: list[Command]) -> int:
        return 1

    def EstimateGetCurrentLocation(self, params: Command.GetCurrentLocation) -> int:
        return 1

    def EstimateMove(self, params: Command.Move) -> int:
        return 1
    
    def EstimateTransfer(self, params: Command.Transfer) -> int:
        return 1
    
    def EstimateJog(self, params: Command.Jog) -> int:
        return 1

    def EstimatePickLid(self, params: Command.PickLid) -> int:
        return 1
    
    def EstimatePlaceLid(self, params: Command.PlaceLid) -> int:
        return 1

    def EstimateLoadWaypoints(self, params: Command.LoadWaypoints) -> int:
        return 1 
    
    def EstimateLoadLabware(self, params: Command.LoadLabware) -> int:
        return 1

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(Pf400Server(), str(args.port))



# tools/common_args.py
import argparse

def get_common_parser(include_port: bool = False) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Galago Tools Manager",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        '--debug',
        action='store_true',
        help='Enable debug mode with detailed logging'
    )
    if include_port:
        parser.add_argument(
            '--port',
            required=True,
            help='Port number for the tool server'
        )
    return parser



from dataclasses import dataclass
import logging
from enum import Enum
from typing import Optional, List
from tools.pf400.tcp_ip import Pf400TcpIp
from tools.base_server import ABCToolDriver
import time 

class RobotError(Enum):
    """Error codes for the PF400 robot"""
    NO_ROBOT = -1009
    POWER_DISABLED = -1046
    POWER_TIMEOUT = -1025
    NOT_HOMED = -1021
    JOINT_OUT_OF_RANGE = -1012

class Axis(Enum):
    """Robot movement axes"""
    X = "x"
    Y = "y"
    Z = "z"
    YAW = "yaw"
    PITCH = "pitch"
    ROLL = "roll"

@dataclass
class RobotConfig:
    """Configuration for the PF400 robot"""
    tcp_host: str
    tcp_port: int
    grasp_plate_buffer_mm: int = 10
    joints: int = 5
    gpl_version: str = "v1"

@dataclass
class Location:
    """Represents a robot location in either joint or Cartesian space"""
    values: List[float]
    
    def to_string(self) -> str:
        """Convert location to space-separated string"""
        return " ".join(str(round(v, 3)) for v in self.values)
    
    @classmethod
    def from_string(cls, loc_string: str) -> 'Location':
        """Create location from space-separated string"""
        values = [float(x) for x in loc_string.split()]
        return cls(values)

class RobotCommunicator:
    """Handles communication with the robot"""
    def __init__(self, tcp_ip: Pf400TcpIp):
        self.tcp_ip = tcp_ip

    def send_command(self, command: str, expected: Optional[str] = None, 
                    timeout: int = 10) -> str:
        """Send command and get response"""
        if expected:
            self.tcp_ip.write_and_expect(command, expected)
            return expected
        return self.tcp_ip.write_and_read(command, timeout=timeout)

    def wait_for_completion(self) -> None:
        """Wait for end of movement signal"""
        self.tcp_ip.wait_for_eom()

    def get_state(self) -> str:
        return self.tcp_ip.write_and_read("sysState")
    
class RobotState:
    """Manages robot state"""
    def __init__(self) -> None:
        self.is_free: bool = False
        self.gripper_axis_override_value: Optional[float] = None

    @property
    def is_plate_gripped(self) -> bool:
        return self.gripper_axis_override_value is not None

class MovementController:
    """Handles robot movement operations"""
    def __init__(self, communicator: RobotCommunicator, state: RobotState, config:RobotConfig):
        self.communicator = communicator
        self.state = state
        self.config = config 

    def move_joints(self, location: Location, profile_id: int) -> None:
        """Move robot using joint coordinates"""
        loc_values = location.values
        if self.state.gripper_axis_override_value is not None:
            loc_values[4] = self.state.gripper_axis_override_value
        if self.config.joints == 5:
            loc_values = loc_values[:5]
        loc_string = Location(loc_values).to_string()
        if self.config.gpl_version == "v1":
            self.communicator.send_command(f"profidx {profile_id}")
            self.communicator.send_command(f"movej {loc_string}")
        elif self.config.gpl_version == "v2":
            self.communicator.send_command(f"movej {profile_id} {loc_string}")
        self.communicator.wait_for_completion()

    def move_cartesian(self, location: Location, motion_profile: int = 1) -> None:
        """Move robot using Cartesian coordinates"""
        loc_values = location.values
        if self.config.joints == 5:
            loc_values = loc_values[:5]
        loc_string = Location(loc_values).to_string()
        if self.config.gpl_version == "v1":
            self.communicator.send_command(f"movec {loc_string}")
        else:
            self.communicator.send_command(f"movec {motion_profile} {loc_string}")
        self.communicator.wait_for_completion()

    def jog(self, axis: Axis, distance: float) -> None:
        """Jog robot along specified axis"""
        current_loc = self._get_current_cartesian_location()
        new_loc = current_loc.values.copy()

        axis_index = {
            Axis.X: 0, Axis.Y: 1, Axis.Z: 2,
            Axis.YAW: 3, Axis.PITCH: 4, Axis.ROLL: 5
        }[axis]
        
        new_loc[axis_index] += distance
        self.move_cartesian(Location(new_loc))

    def _ensure_not_free(self) -> None:
        """Ensure robot is not in free mode"""
        try:
            self._set_free_mode(-1)
            self.state.is_free = False
        except Exception as e:
            raise RuntimeError(f"Failed to unfree arm: {e}")

    def _get_current_cartesian_location(self) -> Location:
        """Get current Cartesian location"""
        response = self.communicator.send_command("wherec")
        return Location([float(x) for x in response.split()[1:-1]])

    def _get_current_joint_location(self) -> Location:
        """Get current joint location"""
        response = self.communicator.send_command("wherej")
        return Location([float(x) for x in response.split()[1:]])

    def _set_free_mode(self, axis: int, timeout: int = 10) -> None:
        """Set free mode for specified axis"""
        self.communicator.send_command(f"freemode {axis}", timeout=timeout)

class GripperController:
    """Handles gripper operations"""
    def __init__(self, communicator: RobotCommunicator, state: RobotState, config: RobotConfig):
        self.communicator = communicator
        self.state = state
        self.config = config

    def grasp_plate(self, plate_width: int, grip_force: int = 10, speed: int = 10) -> None:
        """Grasp a plate"""
        self.state.gripper_axis_override_value = plate_width - self.config.grasp_plate_buffer_mm
        self.communicator.send_command(
            f"graspplate {plate_width} {speed} {grip_force}",
            expected="0 -1"
        )
        self.communicator.wait_for_completion()

    def release_plate(self, plate_width: int, speed: int = 10) -> None:
        """Release a plate"""
        self.communicator.send_command(f"releaseplate {plate_width} {speed}")
        self.communicator.wait_for_completion()

class RobotInitializer:
    """Handles robot initialization"""
    def __init__(self, communicator: RobotCommunicator, config:RobotConfig ):
        self.communicator = communicator
        self.config = config 

    def initialize(self) -> None:
        """Initialize robot with optimized timeouts"""
        try:
            self._ensure_pc_mode()
            self._ensure_power_on()
            self._ensure_robot_attached()
            self._ensure_robot_homed()
        except Exception as e:
            logging.error(f"Initialization failed: {e}")
            raise

    def _ensure_pc_mode(self) -> None:
        """Ensure robot is in PC mode"""
        response = self.communicator.send_command("mode")
        messages = response.split('\n')

        if len(messages) > 1 or messages[0] != "0 0":
            logging.info("Switching to PC mode...")
            response = self.communicator.send_command("mode 0")
            if response != "0":
                raise Exception(f"Could not switch to PC mode: {response}")
            
            response = self.communicator.send_command("mode")
            if response != "0 0":
                raise Exception(f"Could not verify PC mode: {response}")
            
            logging.info("Switched to PC mode")

    def _ensure_power_on(self, target_states:List[str] =["20", "21"], timeout_seconds:int=40) -> None:
        
        start_time = time.time()
        state = self.communicator.get_state()
        self.communicator.send_command("hp 1")
        while True:
            elapsed_time = time.time() - start_time
            if elapsed_time > timeout_seconds:
                raise TimeoutError(f"Timed out waiting for system to reach states {target_states}. Current state: {state}")
            
            response = self.communicator.get_state()
            state = response.split(' ')[1]
            if state in target_states:
                return None
            time.sleep(1)

        
    # def _ensure_power_on_v2(self) -> None:
    #     """Ensure robot power is on with shorter timeout"""
    #     response = self.communicator.send_command("hp")
    #     if response != "0 1":
    #         logging.info("Turning on power...")
    #         response = self.communicator.send_command("hp 1 30")
    #         if response != "0":
    #             raise Exception(f"Could not turn power on: {response}")
            
    #         response = self.communicator.send_command("hp")
    #         if response != "0 1":
    #             raise Exception(f"Could not verify power state: {response}")
            
    #         logging.info("Turned power on")
        
    def _ensure_robot_attached(self) -> None:
        """Ensure robot is attached"""
        response = self.communicator.send_command("attach")
        response_splitted = response.split(" ")
        if response_splitted[0] != "0":
            raise Exception(f"Attach failed. Message is {response}")
        if response != "0 1":
            logging.info("Attaching to robot...")
            response = self.communicator.send_command("attach 1")
            if response != "0":
                raise Exception(f"Could not attach to robot: {response}")
            
            response = self.communicator.send_command("attach")
            if response != "0 1":
                raise Exception(f"Could not verify robot attachment: {response}")
            
            logging.info("Attached to robot")


    def _ensure_robot_homed(self) -> None:
        """Ensure robot is homed by first attempting a move and then homing if needed"""
        # Get current joint location from parent class
        current_loc = self.communicator.send_command("wherej")
        current_joint_loc = " ".join(current_loc.split(" ")[1:])
        
        # Try to move to current position to check if homed
        if self.config.gpl_version == "v1":
            message = self.communicator.send_command(f"movej {current_joint_loc}")
        else:
            message = self.communicator.send_command(f"movej 1 {current_joint_loc}")
        # Robot not homed message (-1021)
        if message == "-1021":
            logging.warning("Homing robot...This may take a moment...")
            message = self.communicator.send_command("home", timeout=40)

            if message != "0":
                raise Exception(f"Got malformed message when homing robot. {message}")

            if self.config.gpl_version == "v1":
                message = self.communicator.send_command(f"movej {current_joint_loc}")
            else:
                message = self.communicator.send_command(f"movej 1 {current_joint_loc}")

            if message != "0":
                raise Exception(f"Could not home robot. {message}")

            logging.info("Robot homed")

class Pf400Driver(ABCToolDriver):
    """Main driver class for the PF400 robot"""
    def __init__(self, tcp_host: str, tcp_port: int, joints:int=5, gpl_version:str="v1") -> None:
        self.state = RobotState()
        self.config = RobotConfig(tcp_host=tcp_host, tcp_port=tcp_port, joints=joints, gpl_version=gpl_version)
        self.tcp_ip: Optional[Pf400TcpIp] = None
        self.communicator: Optional[RobotCommunicator] = None 
        self.gripper: Optional[GripperController] = None
        self.initializer: Optional[RobotInitializer] = None
        self.movement: Optional[MovementController] = None
        
    def initialize(self) -> None:
        """Initialize connection to robot"""
        try:
            # Establish new connection
            self.tcp_ip = Pf400TcpIp(self.config.tcp_host, self.config.tcp_port)
            self.communicator = RobotCommunicator(tcp_ip=self.tcp_ip)
            self.gripper = GripperController(
                communicator=self.communicator,
                state=self.state,
                config=self.config
            )
            self.initializer = RobotInitializer(self.communicator, config=self.config)
            
            # Initialize robot state
            if self.initializer is not None:
                self.initializer.initialize()
            self.movement = MovementController(self.communicator, self.state, self.config)
            logging.info("Successfully connected to PF400")
        except Exception as e:
            logging.error(f"Failed to connect to PF400: {str(e)}")
            # Clean up on failure
            if self.tcp_ip is not None:
                try:
                    self.tcp_ip.close()
                except Exception as close_error:
                    logging.warning(f"Error while closing connection during cleanup: {close_error}")
            self.tcp_ip = None
            self.gripper = None
            self.initializer = None
            self.movement = None
            raise

    def close(self) -> None:
        """Close connection to robot"""
        try:
            if hasattr(self, 'tcp_ip') and self.tcp_ip:
                self.tcp_ip.close()
                self.tcp_ip = None
                logging.info("Successfully closed PF400 connection")
        except Exception as e:
            logging.error(f"Error closing PF400 connection: {str(e)}")
            # Still set to None even if close fails
            self.tcp_ip = None

    # Movement commands
    def movej(self, loc_string: str, motion_profile: int = 1) -> None:
        """Move in joint space"""
        if self.movement is None:
            raise RuntimeError("Robot not initialized")
        self.movement.move_joints(Location.from_string(loc_string), motion_profile)

    def movec(self, loc_string: str, motion_profile: int = 1) -> None:
        """Move in Cartesian space"""
        if self.movement is None:
            raise RuntimeError("Robot not initialized")
        self.movement.move_cartesian(Location.from_string(loc_string), motion_profile)

    def jog(self, axis: str, distance: float) -> None:
        """Jog along specified axis"""
        if self.movement is None:
            raise RuntimeError("Robot not initialized")
        try:
            robot_axis = Axis(axis)
            self.movement.jog(robot_axis, distance)
        except ValueError:
            raise Exception(f"Invalid axis {axis}")

    # Free mode commands
    def free(self) -> None:
        """Enable free mode on all axes"""
        if self.movement is None:
            raise RuntimeError("Robot not initialized")
        logging.info("Freeing robot...")
        self.movement._set_free_mode(0, timeout=15)
        self.state.is_free = True

    def safe_free(self) -> None:
        """Enable free mode on all axes except gripper"""
        if self.movement is None:
            raise RuntimeError("Robot not initialized")
        logging.info("Freeing robot, safe mode (exclude gripper axis 5)...")
        for axis in [1, 2, 3, 4, 6]:
            self.movement._set_free_mode(axis, timeout=15)
        self.state.is_free = True

    def unfree(self) -> None:
        """Disable free mode"""
        if self.movement is None:
            raise RuntimeError("Robot not initialized")
        logging.info("Unfreeing robot...")
        self.movement._set_free_mode(-1, timeout=10)
        self.state.is_free = False

    # Gripper commands
    def graspplate(self, plate_width: int, grip_force: int = 10, speed: int = 10) -> None:
        """Grasp a plate"""
        if self.gripper is None:
            raise RuntimeError("Robot not initialized")
        self.gripper.grasp_plate(plate_width, grip_force, speed)

    def releaseplate(self, plate_width: int, speed: int = 10) -> None:
        """Release a plate"""
        if self.gripper is None:
            raise RuntimeError("Robot not initialized")
        self.gripper.release_plate(plate_width, speed)

    def is_plate_gripped(self) -> bool:
        """Check if plate is currently gripped"""
        return self.state.is_plate_gripped

    # Location commands
    def wherej(self) -> str:
        """Get current joint position"""
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        return self.communicator.send_command("wherej")

    def set_profile_index(self, profile_index:int ) -> None:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        logging.info(f"Setting profile to index {profile_index}")
        self.communicator.send_command(f"profidx {profile_index}")
        profile_current = self.communicator.send_command("profidx")
        logging.info(f"Profile current {profile_current}")

    def wherec(self) -> str:
        """Get current Cartesian position"""
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        return self.communicator.send_command("wherec")

    # Utility commands
    def register_motion_profile(self, profile: str) -> None:
        """Register a motion profile"""
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        logging.info(f"Registering motion profile {profile}...")
        self.communicator.send_command(f"profile {profile}")

    def set_sys_speed(self, speed:int) ->  None:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        self.communicator.send_command(f"mspeed {speed}")

    def get_sys_speed(self) -> str:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        return self.communicator.send_command("mspeed")
    
    def halt(self) -> None:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        self.communicator.send_command("halt")
    
    #If the gripper closed fully then no plate was grabbed/detected.
    def gripper_closed_fully(self) -> None:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        self.communicator.send_command("isfullyclosed")

    def home_all(self) -> None:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        self.communicator.send_command("homeAll")

    def home_if_noplate(self) -> None:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        self.communicator.send_command("homall_ifnoplate")

    def move_to_safe(self) -> None:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        self.communicator.send_command("movetosafe")
    
    def set_gripper_open_position(self, width:float) -> None:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        self.communicator.send_command(f"gripopenpos {width}")

    def get_gripper_open_position(self) -> str:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        return self.communicator.send_command("gripopenpos")

    def set_gripper_close_position(self, width:float) -> None:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")
        self.communicator.send_command(f"gripclosepos {width}")

    def get_gripper_close_position(self) -> str:
        if self.communicator is None:
            raise RuntimeError("Robot not initialized")        
        return self.communicator.send_command("gripclosepos")

    def __del__(self) -> None:
        """Cleanup when driver is destroyed"""
        self.close()

# if __name__ == "__main__":
#     driver = Pf400Driver(
#         tcp_host="192.168.0.1",
#         tcp_port=10100,
#         joints=6,
#         gpl_version="v2"
#     )
#     driver.initialize()

#     logging.info("Getting system speed")
#     driver.get_sys_speed()
#     logging.info("Getting close width")
#     driver.get_gripper_close_position()
#     logging.info("Getting open width")
#     driver.get_gripper_open_position()




from tools.toolbox.db import Db 
from typing import Any, Dict, Optional

db = Db()

def get_inventory(name:str) -> Any:
    response = db.get_by_id_or_name(name, "inventory")
    return response
    
def get_all_inventory(workcell_name: Optional[str] = None) -> Any:
    """
    Get inventory items, optionally filtered by workcell_name.
    If workcell_name is None, attempts to get all inventory items.
    """
    if workcell_name:
        response = db.get_data(f"inventory?workcell_name={workcell_name}")
    else:
        response = db.get_data("inventory?all_workcells=true")
    return response

# Nest functions
def get_nests(workcell_name: str) -> Any:
    response = db.get_data(f"nests?workcell_name={workcell_name}")
    return response

def get_nest(nest_id: int) -> Any:
    response = db.get_by_id_or_name(nest_id, "nests")
    return response

def create_nest(nest_data: Dict[str, Any]) -> Any:
    response = db.post_data(nest_data, "nests")
    return response

def update_nest(nest_id: int, nest_data: Dict[str, Any]) -> Any:
    response = db.update_data(nest_id, nest_data, "nests")
    return response

def delete_nest(nest_id: int) -> Any:
    response = db.delete_data(nest_id, "nests")
    return response

# Plate functions
def get_plates(workcell_name: str) -> Any:
    response = db.get_data(f"plates?workcell={workcell_name}")
    return response

def get_plate(plate_id: int) -> Any:
    response = db.get_by_id_or_name(plate_id, "plates")
    return response

def get_plate_info(plate_id: int) -> Any:
    response = db.get_data(f"plates/{plate_id}/info")
    return response

def create_plate(plate_data: Dict[str, Any]) -> Any:
    response = db.post_data(plate_data, "plates")
    return response

def update_plate(plate_id: int, plate_data: Dict[str, Any]) -> Any:
    response = db.update_data(plate_id, plate_data, "plates")
    return response

def delete_plate(plate_id: int) -> Any:
    response = db.delete_data(plate_id, "plates")
    return response

# Well functions
def get_wells(plate_id: int) -> Any:
    response = db.get_data(f"wells?plate_id={plate_id}")
    return response

# Reagent functions
def get_reagents(plate_id: int) -> Any:
    response = db.get_data(f"reagents?plate_id={plate_id}")
    return response

def get_workcell_reagents(workcell_name: str) -> Any:
    response = db.get_data(f"reagents?workcell_name={workcell_name}")
    return response

def create_reagent(reagent_data: Dict[str, Any]) -> Any:
    response = db.post_data(reagent_data, "reagents")
    return response

def update_reagent(reagent_id: int, reagent_data: Dict[str, Any]) -> Any:
    response = db.update_data(reagent_id, reagent_data, "reagents")
    return response

def delete_reagent(reagent_id: int) -> Any:
    response = db.delete_data(reagent_id, "reagents")
    return response

def get_reagents_by_name_and_quantity(reagent_name: str, quantity: int, workcell_name: str) -> Any:
    """
    Returns an array of reagents with the specified name and quantity from a single plate.
    
    Args:
        reagent_name: The name of the reagent to search for
        quantity: The required quantity of the reagent
        workcell_name: The name of the workcell to search in
        
    Returns:
        A list of matching reagents or raises an exception if none found
    """
    # Get all reagents for the workcell
    all_reagents = get_workcell_reagents(workcell_name)
    
    # Filter reagents by name and quantity
    matching_reagents = [r for r in all_reagents if r.get("name") == reagent_name and r.get("quantity", 0) >= quantity]
    
    if not matching_reagents:
        raise Exception(f"No reagents found with name '{reagent_name}' and quantity {quantity} in workcell '{workcell_name}'")
    
    # Group reagents by plate_id
    reagents_by_plate = {}
    for reagent in matching_reagents:
        plate_id = reagent.get("plate_id")
        if plate_id not in reagents_by_plate:
            reagents_by_plate[plate_id] = []
        reagents_by_plate[plate_id].append(reagent)
    
    # Find the plate with the most matching reagents
    best_plate_id = None
    max_reagents = 0
    for plate_id, reagents in reagents_by_plate.items():
        if len(reagents) > max_reagents:
            max_reagents = len(reagents)
            best_plate_id = plate_id
    
    if best_plate_id is None:
        raise Exception(f"Could not determine a suitable plate for reagent '{reagent_name}' with quantity {quantity}")
    
    return reagents_by_plate[best_plate_id]
 




import logging
import os
import argparse
import pythoncom
import threading
import time
from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.bravo_pb2 import Command, Config
from .driver import VPrepDriver, kill_vworks
from typing import Callable, Any 

_thread_local = threading.local()

class VPrepServer(ToolServer):
    toolType = "vprep"
    config: Config
    
    def __init__(self) -> None:
        super().__init__()
        self.main_thread_id = threading.get_ident()
        logging.info(f"VPrepServer initialized in thread ID: {self.main_thread_id}")
        
        # Initialize COM in the main thread
        if os.name == "nt":
            pythoncom.CoInitialize()
            logging.info("COM initialized in main thread")
        
        # Flag to signal server shutdown
        self.shutdown = False
    
    def _configure(self, request: Config) -> None:
        logging.info(f"Configuring VPrep in thread ID: {threading.get_ident()}")
        self.config = request
        logging.info("VPrep configuration complete")
    
    def _get_thread_driver(self, force_new:bool=False) -> Any:

        thread_id = threading.get_ident()
        
        # Check if we need to initialize COM for this thread
        if not hasattr(_thread_local, 'com_initialized'):
            logging.info(f"Initializing COM in thread ID: {thread_id}")
            pythoncom.CoInitialize()
            _thread_local.com_initialized = True
        
        # Create or verify existing driver
        try:
            # If we need a new driver or don't have one
            if force_new or not hasattr(_thread_local, 'driver'):
                # If we had a previous driver, try to close it properly
                if hasattr(_thread_local, 'driver'):
                    try:
                        logging.info(f"Closing previous driver in thread ID: {thread_id}")
                        _thread_local.driver.close()
                    except Exception as e:
                        logging.warning(f"Error closing previous driver: {e}")
                
                # Force kill VWorks to ensure clean state
                kill_vworks()
                time.sleep(0.5)
                
                # Create new driver
                logging.info(f"Creating new VPrepDriver in thread ID: {thread_id}")
                _thread_local.driver = VPrepDriver(init_com=False)
                _thread_local.driver.login()
                logging.info(f"VPrepDriver created and logged in for thread ID: {thread_id}")
            else:
                # Verify the existing driver is still responsive
                try:
                    # Try to pump messages which will fail if COM is disconnected
                    pythoncom.PumpWaitingMessages()
                except Exception as e:
                    logging.warning(f"COM connection test failed, recreating driver: {e}")
                    return self._get_thread_driver(force_new=True)
            
            return _thread_local.driver
            
        except Exception as e:
            logging.error(f"Failed to create VPrepDriver in thread {thread_id}: {e}")
            # Clean up and try again with a fresh environment
            if hasattr(_thread_local, 'driver'):
                try:
                    _thread_local.driver.close()
                except Exception as e:
                    logging.warning(f"Failed to close VPrepDriver: {e}")
                delattr(_thread_local, 'driver')
            
            # If this is already a retry, give up to avoid infinite recursion
            if force_new:
                raise RuntimeError(f"Failed to initialize VPrep driver after recovery attempt: {str(e)}")
            
            # Try once more with force_new=True
            logging.info("Attempting recovery by creating a fresh driver")
            kill_vworks()  # Make sure VWorks is killed
            time.sleep(1)  # Give more time for cleanup
            return self._get_thread_driver(force_new=True)
    
    def cleanup(self) -> None:
        """Clean up resources properly for all threads"""
        logging.info(f"Cleanup called from thread ID: {threading.get_ident()}")
        
        # Signal shutdown
        self.shutdown = True
        
        # Clean up the driver in the current thread if it exists
        if hasattr(_thread_local, 'driver'):
            try:
                logging.info(f"Cleaning up driver for thread ID: {threading.get_ident()}")
                _thread_local.driver.close()
                delattr(_thread_local, 'driver')
            except Exception as e:
                logging.error(f"Error closing driver: {e}")
        
        # Clean up COM in the current thread if it was initialized
        if hasattr(_thread_local, 'com_initialized'):
            try:
                logging.info(f"Uninitializing COM in thread ID: {threading.get_ident()}")
                pythoncom.CoUninitialize()
                delattr(_thread_local, 'com_initialized')
            except Exception as e:
                logging.error(f"Error uninitializing COM: {e}")
        
        # Make sure VWorks is killed
        kill_vworks()
    
    def _execute_with_retry(
        self,
        operation_name: str,
        operation_func: Callable[[VPrepDriver], None],
        max_retries: int = 2
    ) -> None:
        """Execute an operation with automatic retry on RPC errors"""
        retries = 0
        last_error = None
        while retries <= max_retries:
            try:
                if retries > 0:
                    logging.info(f"Retry #{retries} for {operation_name}")
                    # If we're retrying, force a new driver
                    driver = self._get_thread_driver(force_new=True)
                else:
                    # First attempt uses existing or new driver
                    driver = self._get_thread_driver()
                
                # Process COM messages before running
                if os.name == "nt":
                    pythoncom.PumpWaitingMessages()
                
                # Run the operation
                return operation_func(driver)
                
            except Exception as e:
                last_error = e
                error_str = str(e)
                
                # Check if this is an RPC error
                if "RPC server is unavailable" in error_str or "marshalled for a different thread" in error_str:
                    logging.warning(f"RPC connection error in {operation_name}: {e}")
                    retries += 1
                    # Force kill VWorks
                    logging.info("RPC unavailable killing vworks")
                    kill_vworks()
                    time.sleep(1)  # Wait before retry
                else:
                    # This is not an RPC error, don't retry
                    logging.error(f"Non-RPC error in {operation_name}: {e}")
                    raise
        
        # If we exhausted all retries
        logging.error(f"Failed {operation_name} after {max_retries} retries")
        raise last_error
    
    def RunProtocol(self, params: Command.RunProtocol) -> None:
        thread_id = threading.get_ident()
        logging.info(f"RunProtocol called from thread ID: {thread_id}")
        
        def run_op(driver:VPrepDriver) -> None:
            return driver.run_protocol(params.protocol_file)
        
        self._execute_with_retry("RunProtocol", run_op)
    
    def RunRunset(self, params: Command.RunRunset) -> None:
        thread_id = threading.get_ident()
        logging.info(f"RunRunset called from thread ID: {thread_id}")
        def run_op(driver: VPrepDriver) -> None:
            return driver.run_runset(params.runset_file)
        
        self._execute_with_retry("RunRunset", run_op)
   
    def EstimateRunProtocol(self, params: Command.RunProtocol) -> int:
        return 60  # Return a more realistic estimate in seconds
    
    def EstimateRunRunset(self, params: Command.RunRunset) -> int:
        return 120  # Return a more realistic estimate in seconds
    
if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(VPrepServer(), str(args.port))



import serial
import time
import logging
import typing as t

from tools.base_server import ABCToolDriver

ERROR_CODES = {
            "101": "DC motor controller error.",
            "102": "Error caused by speed failure (e.g mechanical locking).",
            "103": "Shaker not initialized or initialization parameters were incorrect after powering on.",
            "104": "Initialization error.",
            "105": "Shaker did not home after stop command.",
            "201": "Error due to failed response from temperature sensors or incorrect internal settings of temperature sensors.",
            "202": "Temperature communication bus error.",
            "203": "Sensor with the requested ID was not found.",
            "204": "Errors caused by a faulty temperature measurement during run.",
            "300": "General error.",
            "301": "IC-Driver error.",
            "303": "Verification error by the unlock position. Please RESTART the system.",
            "304": "Error caused by unsuccessful reach the lock position (timeout). Please RESTART the system.",
            "305": "Error caused by unsuccessful reach the unlock position (timeout). Please RESTART the system.",
            "306": "Error caused by unsuccessful reach the lock position (over current). Please RESTART the system.",
            "307": "Error caused by unsuccessful reach the unlock position (over current). Please RESTART the system.",
        }

SHAKE_STATES = {
                "0": "Shaking",
                "1": "Shake Cycle Stopped and Homed",
                "2": "Shake Cycle Emergency Stop. Re-home shaker",
                "3": "Home",
                "4": "Manual Mode",
                "5": "Accelerating",
                "6": "Decelerate",
                "7": "Decelerate to Stop.",
                "90": "ECO Mode.",
                "99": "Boot Process Running.",
            }

class BioshakeDriver(ABCToolDriver):
    def __init__(self, port: str):
        self.port = port
        self.ser = serial.Serial(port, baudrate=9600, timeout=3)
        self.shake_state = None

    def __del__(self) -> None:
        self.disconnect()

    def _send_command(self, command: str) -> str:
        if not self.ser.is_open:
            self.ser.open()

        full_command = command + "\r\n"
        self.ser.write(full_command.encode("ascii"))
        response = self.ser.readline().decode("ascii").strip()
        if response == "e":
            error_list = self.get_error_list()
            if not error_list:
                error_list = "Unknown error - no error codes returned from device"
            error_msg = f"BioShake error during command '{command}': {error_list}"
            logging.error(error_msg)
            raise Exception(error_msg)
        return response

    """
    Initialization commands 
    """

    def connect(self) -> None:
        if self.ser is not None and self.ser.is_open:
            logging.warning("Already connected to the device...")
            return None
        try:
            self.ser = serial.Serial(port=self.port, baudrate=9600, timeout=1)
        except serial.SerialException as e:
            logging.error(f"Error opening serial port: {e}")
            raise

    def disconnect(self) -> None:
        if self.ser.is_open:
            self.ser.close()

    def reset(self) -> None:
        self._send_command("reset")
        attempts: int = 0
        while attempts < 5:
            try:
                if self.ser:
                    self.disconnect()
                    logging.info("Device reset successfully.")
                    self.ser.open()
                    break
                else:
                    self.connect()
                    break
            except serial.SerialException:
                attempts += 1
                time.sleep(1)
        if attempts == 5:
            logging.error("Failed to reset device after multiple attempts.")    
            raise Exception("Failed to reset device.")

    def get_version(self) -> str:
        version = self._send_command("v")
        if version is None:
            logging.error("Failed to get version.")
            return "Unknown"
        return version.strip()

    def get_error_list(self) -> str:
        result = self._send_command("gel").strip()
        errors = []
        for err in result.split(";"):
            code = err.strip().strip("{}")
            if code in ERROR_CODES:
                errors.append(ERROR_CODES[code])
            else:
                errors.append(code)
        error_message = "\n".join(errors)
        logging.error("Error list: " + error_message)
        return error_message

    def flash_led(self) -> None:
        self._send_command("fled")

    """Shake Commmands"""
    def home(self) -> None:
        if not self.is_gripper_closed():
            self.grip()
        self._send_command("sgh")
        while self._get_shake_state() != "Home":
            time.sleep(0.5)

    def stop_shake(self) -> None:
        self._send_command("soff")
        while self._get_shake_state() != "Home":
            time.sleep(0.5)
        self.ungrip()

    def _set_shake_speed(self, speed: int) -> None:
        if 0 < speed < 9999:
            self._send_command("ssts" + str(speed))
        else:
            raise ValueError("Please enter a valid target speed. (Non-negative and no larger than 4 digits.)")

    def _set_acceleration(self, acceleration: int) -> None:
        if 0 < acceleration <= 99:
            self._send_command("ssa" + str(acceleration))
        else:
            logging.error("Please enter an acceleration value within a 1 to 2 digit range.")

    def _get_remaining_time(self) -> int:
        response = self._send_command("gsrt")
        try:
            return int(response)
        except ValueError:
            logging.error("Invalid remaining time received.")
            return 0

    def _get_shake_state(self) -> str:

        result = self._send_command("gsst")
        if result in SHAKE_STATES:
            return SHAKE_STATES[result]
        elif result is None:
            return "Unable to return Shaker status"
        return result

    def _get_target_speed(self) -> int:
        response = self._send_command("gsts")
        try:
            return int(response)
        except ValueError:
            logging.error("Invalid target speed received.")
            return 0

    def _get_speed_actual(self) -> int:
        response = self._send_command("gsas")
        try:
            return int(response)
        except ValueError:
            logging.error("Invalid actual speed received.")
            return 0

    def _get_acceleration_actual(self) -> int:
        response = self._send_command("gsa")
        try:
            return int(response)
        except ValueError:
            logging.error("Invalid acceleration value received.")
            return 0



    """Gripper Commmands"""
    def is_gripper_closed(self) -> bool:
        elm_state = self.get_elm_state_as_string()
        if elm_state.lower() == "elmlocked":
            return True
        else:
            return False

    def grip(self) -> None:
        self._send_command("selp")

    def ungrip(self) -> None:
        self._send_command("seup")

    def start_shake(self, speed: t.Optional[int] = None, acceleration: t.Optional[int] = None
    ) -> None:
        if not self.is_gripper_closed():
            self.grip()
        if speed is not None:
            self._set_shake_speed(speed)
        if acceleration is not None:
            self._set_acceleration(acceleration)
        
        self._send_command("son")

    def get_elm_state_as_string(self) -> str:
        return self._send_command("gesas")

    def set_elm_lock_pos(self) -> None:
        self.grip()

    def set_elm_unlock_pos(self) -> None:
        self.ungrip()
    
    def wait_for_shake(self, timeout: int) -> None:
        start_time = time.time()
        while time.time() - start_time < timeout:
            shake_state = self._get_shake_state()
            if shake_state == "Home":
                logging.info("Shake finished")
                return None
            time.sleep(0.5)
        
        raise TimeoutError(f"Shake operation timed out after {timeout} seconds")
    
    
    def set_shake_target_speed(self, rpm: int) -> None:
        self._set_shake_speed(rpm)

    def shake_on_with_runtime(
        self, seconds: int, speed: t.Optional[int] = None, acceleration: t.Optional[int] = None
    ) -> None:
        if not self.is_gripper_closed():
            self.grip()
        if speed is not None:
            self._set_shake_speed(speed)
        if acceleration is not None:
            self._set_acceleration(acceleration)
        else:
            self._set_acceleration(10)
        self._send_command("sonwr" + str(seconds))
        while self._get_shake_state() != "Home":
            time.sleep(0.5)
        self.ungrip()

    def get_shake_remaining_time(self) -> int:
        return self._get_remaining_time()


    """Temperature Commmands"""
    def temp_on(self) -> None:
        self._send_command("ton")

    def temp_off(self) -> None:
        self._send_command("toff")

    def set_tmp(self, temp: int) -> None:
        if 0 < temp < 990:
            self._send_command("stt" + str(temp))
        else:
            logging.error(f"Enter a valid temperature target {temp}. Must be a three digit number in the range 0-999.")




import serial
import time
import logging
from tools.base_server import ABCToolDriver
import threading 
from typing import Optional, Union
from tools.app_config import Config 


ERROR_CODES = {
    "06163": "Failed to load plate. There might be a plate at specified location."
}

def try_ascii_decode(data: Union[str,bytes]) -> str:
    if isinstance(data, str):
        return data
    data_string = ""
    try:
        data_string = data.decode("ascii")
    except Exception:
        raise serial.SerialException(f"error decoding to ascii string: {data!r}")
    return data_string


def serial_read(serial_port: serial.Serial) -> str:
    reply = serial_port.read_until(expected=b"\n")
    reply_string = try_ascii_decode(reply)
    if reply_string == "":
        logging.warning("Liconic STX returned empty response")
        return ""
    # Ignore any malformed responses for now.
    if not (reply_string[-1] == "\n" and reply_string[-2] == "\r"):
        logging.warning(f"Liconic STX returned malformed response {reply_string}")
        return ""
    return reply_string[0:-2]


WAIT_TIMEOUT = 90

class LiconicStxDriver(ABCToolDriver):
    # Example: com_port can be "COM1"
    def __init__(self, com_port: str) -> None:
        self.config = Config()
        self.config.load_workcell_config()
        # if self.config.app_config.data_folder:
        #     self.co2_log_path = os.path.join(self.config.app_config.data_folder,"sensors","liconic")
        self.serial_port = serial.Serial(
            com_port, 9600, timeout=1, parity=serial.PARITY_EVEN
        )
        self.lock = threading.Lock()
        self.monitor_thread: Optional[threading.Thread] = None
        self.is_busy = False
        self.monitoring = False
        self.co2_out_of_range = False
        self.connect()

    def write(self, message: str) -> None:
        with self.lock:
            try:
                self.serial_port.write((message + "\r").encode("ascii"))
            except Exception as e:
                raise RuntimeError(f"Failed to communicate with liconic. {e}")

    # Read a single message from the serial port.
    # strict - if true, exceptions will be thrown if read fails.
    # max-attempts - number of times to retry if an empty string or error message is received.
    def read(self, strict: bool = True, max_attempts: int = 25) -> str:
        attempt = 0
        # The instruments will sometimes return an empty string or "{}" even during normal operation.
        while attempt < max_attempts:
            # Read a message.
            with self.lock:
                reply_string = serial_read(self.serial_port)

            if reply_string == "":
                attempt += 1
                logging.info(f"Retrying...Attempt {attempt}")
                continue

            # If we have reached here, the message is valid. Return it.
            return reply_string

        # If we have reached here, we have exceeded max attempts.
        # The instrument may have disconnected.
        if strict:
            raise serial.SerialException(
                f"No valid data was received. {self.serial_port.name} may have disconnected."
            )
        return ''

    def wait_for_ready(self, timeout: int = WAIT_TIMEOUT, custom_error: Optional[str]=None) -> None:
        times = 0
        while True:
            self.write("RD 1915")
            ready = self.read()
            if ready == "1":
                logging.info(f"Waited for {times} seconds")
                return
            if self.has_error():
                error_code = self.get_error_code()
                if error_code in ERROR_CODES:
                    raise Exception(f"{ERROR_CODES[error_code]}")
                else:
                    raise Exception(f"Liconic has errored with code {error_code}")
            logging.info("Liconic stx is busy. Waiting...")
            times += 1
            if times > timeout:
                raise Exception(custom_error or "Liconic has timed out waiting for command")
            time.sleep(1)

    def has_error(self) -> bool:
        self.write("RD 1814")
        return self.read() == "1"

    def get_error_code(self) -> str:
        self.write("RD DM200")
        return self.read()

    def expect_response(self, expected: str, strict: bool = False) -> None:
        response = self.read()
        if response != expected:
            if strict:
                raise Exception(f"Expected response {expected}. Got {response}")
            else:
                logging.warning(f"Expected response {expected}. Got {response}")

    def log(self, msg: str) -> None:
        logging.info(f"liconic_stx_legacy_driver: {msg}")

    def connect(self) -> None:
        self.log("Connecting...")
        self.write("CR")
        self.expect_response("CC")

    def close(self) -> None:
        self.log("Closing...")
        self.write("CQ")
        self.expect_response("CF")
        self.serial_port.close()

    def reset(self) -> None:
        self.log("Resetting...")
        self.write("ST 1900")
        self.expect_response("OK")
        self.wait_for_ready()

    def initialize(self) -> None:
        self.log("Initializing...")
        self.write("ST 1801")
        self.expect_response("OK")
        self.wait_for_ready()

    def load_plate(self, cassette: int, level: int) -> None:
        logging.info("Loading plate")
        self.is_busy = True
        try:
            self.log(f"Loading plate (cassette={cassette}, level={level})...")
            self.write(f"WR DM0 {cassette}")
            self.expect_response("OK")
            self.write(f"WR DM5 {level}")
            self.expect_response("OK")
            self.write("ST 1904")
            self.expect_response("OK")

            self.wait_for_ready()
            self.check_shovel_station_sensor("0")
            # Verify that the plate is not on the transfer station any longer.
            self.check_transfer_station_sensor("0")
        except Exception as e:
            raise RuntimeError(f"{e}")
        finally:
            self.is_busy = False
            logging.info("Loading plate complete")

    def unload_plate(self, cassette: int, level: int) -> None:
        self.is_busy = True
        try:
            self.log(f"Unloading plate (cassette={cassette}, level={level})...")
            self.write(f"WR DM0 {cassette}")
            self.expect_response("OK")
            self.write(f"WR DM5 {level}")
            self.expect_response("OK")
            self.write("ST 1905")
            self.expect_response("OK")

            self.wait_for_ready()
            self.check_shovel_station_sensor("0")
            # Verify that the plate is on the transfer station.
            self.check_transfer_station_sensor("1")
        except Exception as e:
            raise RuntimeError(f"{e}")
        finally:
            self.is_busy = False

    def read_error_code(self) -> None:
        self.write("RD DM200")
        response = self.read()
        logging.info(f"Current error code {response}")

    def check_transfer_station_sensor(self, expected: str = "0") -> None:
        self.write("RD 1813")
        self.expect_response(expected)

    def check_shovel_station_sensor(self, expected: str = "0") -> None:
        self.write("RD 1812")
        self.expect_response(expected)

    def show_cassette(self, cassette: int) -> None:
        rotate_to_cassette = (cassette - 3) % 10
        self.write(f"WR DM0 {rotate_to_cassette}")
        self.expect_response("OK")
        self.wait_for_ready()

    def raw(self, message: str) -> None:
        self.write(message)
        response = self.read()
        logging.info(f"response: {response}")

    def get_co2_set_point(self) -> str:
        self.write("RD DM894")
        return self.read()

    def set_co2_set_point(self, level: float) -> None:
        self.write(f"WR DM894 {str(int(level * 100)).zfill(5)}")
        self.expect_response("OK")

    def get_co2_cur_level(self) -> str:
        self.write("RD DM984")
        return self.read()

    def start_monitor(self) -> None:
        self.monitor_thread = threading.Thread(target=self.monitor_co2_level)
        self.monitor_thread.daemon = True
        self.monitor_thread.start()
        return None

    # def write_co2_log(self, value:str) -> None:

    #     #Write to local files
    #     if self.co2_log_path is None:
    #         return
    #     if(os.path.exists(self.co2_log_path) is False):
    #         logging.debug("folder does not exist. creating folder")
    #         os.mkdir(self.co2_log_path )
    #     today =  datetime.today().strftime('%Y-%m-%d')
    #     today_folder = os.path.join(self.co2_log_path ,today)
    #     if(os.path.exists(today_folder) is False):
    #         logging.debug("folder does not exist. creating folder")
    #         os.mkdir(today_folder)
    #     trace_file = os.path.join(today_folder, "liconic_co2.txt")
    #     try:
    #         if os.path.exists(trace_file) is False:
    #             with open(trace_file, 'w+') as f:
    #                 f.write('Time,Value\n')
    #     except Exception as e:
    #         logging.debug(e)
    #         return
        
    #     try:
    #         with open(trace_file, 'a') as f:
    #             f.write(f"{datetime.today()},{value}\n")
    #     except Exception as e:
    #         logging.debug(e)
    #         return

    # def check_last_co2_level(self, data_points:int) -> Optional[int]:
    #     if self.co2_log_path is None:
    #         return
    #     today =  datetime.today().strftime('%Y-%m-%d')
    #     today_file = os.path.join(self.co2_log_path,today,"liconic_co2.txt")
    #     logging.info(today_file)
    #     if(os.path.exists(today_file) is False):
    #         logging.debug("folder does not exist.")
    #         return None 
    #     with open(today_file, "r") as f:
    #         lines = f.readlines()
    #         logging.info(F"length of lines {len(lines)}")
    #         return None

    def monitor_co2_level(self) -> None:
        self.monitoring = True
        try:
            while True:
                if not self.is_busy:
                    co2_level = float(self.get_co2_cur_level())/100
                    self.write_co2_log(str(co2_level))
                    if co2_level < 3:
                        self.co2_out_of_range = True
                        error_message = f"CO2 level is low: {co2_level}%"
                        logging.warning(error_message)
                        self.co2_out_of_range = False
                time.sleep(300)
                
        except Exception as e:
            logging.warning("Liconic monitor thread encounter an error. Restart the tool")
            logging.exception("Error is" + str(e))
            self.monitoring = False
        return None


import os
from setuptools import setup
import subprocess
from setuptools.command.build_py import build_py as _build_py
import shutil 
import sys 


# Version is now defined in tools/version.py and imported here.
# Scripts that need to extract the version should use:
# grep -oP "__version__ = ['\"]([^'\"]+)" tools/version.py
sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
from tools.version import __version__


base = None
if sys.platform == "win32":
    base = "Win32GUI"  # Use "Console" if you want a console window 


class BuildProtobuf(_build_py):
    def run(self) -> None:
        base_dir = os.path.dirname(os.path.abspath(__file__))  # Root directory
        proto_src = os.path.join(base_dir, "interfaces")
        grpc_interfaces_output_dir = os.path.join(base_dir, "tools", "grpc_interfaces")  

        os.makedirs(grpc_interfaces_output_dir, exist_ok=True)

        grpc_proto_dir = os.path.join(proto_src, "tools", "grpc_interfaces")
        if not os.path.exists(grpc_proto_dir):
            raise FileNotFoundError(f"Expected proto directory {grpc_proto_dir} not found!")

        grpc_proto_files = [
            os.path.join(grpc_proto_dir, proto_file)
            for proto_file in os.listdir(grpc_proto_dir)
            if proto_file.endswith(".proto")
        ]

        root_proto_files = [
            os.path.join(proto_src, proto_file)
            for proto_file in os.listdir(proto_src)
            if proto_file.endswith(".proto")
        ]

        if grpc_proto_files:
            subprocess.run(
                [
                    "python", "-m", "grpc_tools.protoc",
                    f"-I{proto_src}",
                    f"--python_out={grpc_interfaces_output_dir}/",
                    f"--pyi_out={grpc_interfaces_output_dir}/",
                    f"--grpc_python_out={grpc_interfaces_output_dir}/",
                    *grpc_proto_files,
                ],
                check=True,
            )

        grpc_generated_path = os.path.join(grpc_interfaces_output_dir, "tools", "grpc_interfaces")
        if os.path.exists(grpc_generated_path):
            for file in os.listdir(grpc_generated_path):
                if file.endswith(".py") or file.endswith(".pyi"):
                    shutil.move(
                        os.path.join(grpc_generated_path, file),
                        os.path.join(grpc_interfaces_output_dir, file),
                    )

        if root_proto_files:
            subprocess.run(
                [
                    "python", "-m", "grpc_tools.protoc",
                    f"-I{proto_src}",
                    f"--python_out={grpc_interfaces_output_dir}/",
                    f"--pyi_out={grpc_interfaces_output_dir}/",
                    f"--grpc_python_out={grpc_interfaces_output_dir}/",
                    *root_proto_files,
                ],
                check=True,
            )

        super().run()

def readme() -> str:
    """
    Use PYPI.md for PyPI documentation if it exists,
    otherwise fall back to README.md
    """
    pypi_path = os.path.join(os.path.dirname(__file__), "PYPI.md")
    if os.path.exists(pypi_path):
        return open(pypi_path).read()
    
    # Fallback to README.md
    readme_path = os.path.join(os.path.dirname(__file__), "README.md")
    if os.path.exists(readme_path):
        return open(readme_path).read()
    return ""

def read_requirements(filename:str) -> list[str]:
    requirements_path = os.path.join(os.path.dirname(__file__), filename)
    print(requirements_path)
    with open(requirements_path) as f:
        return [line.strip() for line in f
                if line.strip() and not line.startswith('#')]

def find_tool_packages() -> list[str]:
    """Find all packages and subpackages in the current directory"""
    packages = []
    for root, dirs, files in os.walk('tools'):  # Only look inside `tools/`
        if '__init__.py' in files:
            package_path = root.replace('/', '.')
            print(f"Detected package: {package_path}")
            packages.append(package_path)

    # Ensure grpc_interfaces is included if it exists
    if os.path.exists("tools/grpc_interfaces") and "__init__.py" in os.listdir("tools/grpc_interfaces"):
        packages.append("tools.grpc_interfaces")

    return packages

setup(
    name='galago-tools',
    version=__version__, # latest version
    packages=find_tool_packages(),
    package_dir={'': '.'},
    license='Apache-2.0',  # Standard SPDX identifier
    description='Open Source Lab Automation GRPC Library',
    long_description=readme(),
    install_requires=read_requirements('requirements.txt'),
    include_package_data=True,
    package_data={'tools': ["site_logo.png",
                            'vcode/deps/*.dll',
                            'plateloc/deps/*.dll',
                            'vspin/deps/*.dll',
                            'hig_centrifuge/deps/*.dll',
                            'bravo/deps/*.dll',
                            'minihub/deps/*.dll',
                            "favicon.ico",
                            'grpc_interfaces/*.py']},
    url='https://github.com/sciencecorp/galago-tools',
    author='Science Corporation',
    python_requires=">=3.9",
    author_email='',
    long_description_content_type="text/markdown",
    entry_points={
        'console_scripts': [
            'galago=tools.cli:main',
            'galago-serve=tools.cli:serve',
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",  # Standard classifier for Apache
    ],
    cmdclass={
        'build_py': BuildProtobuf
    },
)


import os
import sys
import argparse
import logging
import subprocess

# Move heavy imports inside functions to delay loading
def get_shell_command(tool: str, file: str) -> list[str]:
    if tool:
        return [sys.executable, '-m', f'tools.{tool}.server']
    elif file:
        return [sys.executable, file]
    else:
        raise RuntimeError("Either tool or file must be provided.")

def serve() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('--port', required=True, help="Port must be provided.")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--tool', help="Name of the tool to run (e.g., 'xpeel').")
    group.add_argument('--file', help="Path to the file to run.")
    
    known, remaining = parser.parse_known_args()
    use_shell = os.name == 'nt'
    sys.argv = [sys.argv[0]] + remaining
    command = get_shell_command(known.tool, known.file)
    command.append(f'--port={known.port}')
    
    # Set the PYTHONPATH to include the project root.
    env = os.environ.copy()
    project_root = os.getcwd()
    current_pythonpath = env.get("PYTHONPATH", "")
    env["PYTHONPATH"] = project_root + (os.pathsep + current_pythonpath if current_pythonpath else "")
    
    subprocess.Popen(command, shell=use_shell, universal_newlines=True, env=env)

def main() -> None:
    logging.basicConfig(level=logging.DEBUG)
    
    # Define top-level arguments that you want to process
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--help", action="help", help="Show this help message and exit")
    parser.add_argument("--discover", action="store_true", help="Autodiscover tools")
    parser.add_argument("--console", action="store_true", help="Launch in console mode")
    
    # Parse known arguments and get the remaining arguments (if any)
    known, remaining = parser.parse_known_args()
    
    # Remove the consumed arguments from sys.argv
    sys.argv = [sys.argv[0]] + remaining
    
    if known.console:
        from tools.launch_console import main as launch_console_main
        sys.exit(launch_console_main())
    elif known.discover:
        from tools.discover_tools import main as autodiscover_main
        sys.exit(autodiscover_main())
    else:
        from tools.launch_tools import main as launch_tools_main
        sys.exit(launch_tools_main())

if __name__ == "__main__":
    main()


import logging
import time
from tools.base_server import ABCToolDriver
import serial

class ALPS3000Driver(ABCToolDriver):
    def __init__(self, profile: str, port: str, baudrate: int = 9600, timeout: int = 1) -> None:
        self.profile: str = profile
        self.live: bool = False
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout
        self.client : serial.Serial
        self.initialize()

    def initialize(self) -> None:
        self.client = serial.Serial(self.port, self.baudrate, timeout=self.timeout)
        return None
    
    def send_command(self, command: str) -> None:
        self.client.write(f"{command}\r".encode())
        time.sleep(0.1)
        try:
            response = self.client.readline().decode().strip()
            if response != "ok" and response != "ff":
                raise RuntimeError(f"Failed to execute command: {response}")
        except RuntimeError as e:
            self.live = False
            logging.error(f"Failed to execute command {str(e)}")
        
    def close(self) -> None:
        if self.client:
            self.client.close()

    def get_status(self) -> None:
        return self.send_command("?")

    def seal_plate(self) -> None:
        self.send_command("S")

    def get_error(self) -> None:
        return self.send_command("E")

    def set_sealing_temperature(self, temperature: int) -> None:
        if 0 <= temperature <= 999:
            return self.send_command(f"A{temperature:03d}")
        else:
            raise IndexError("Temperature out of range ")

    def set_sealing_time(self, time_10ths: int) -> None:
        if 0 <= time_10ths <= 99:
            self.send_command(f"B{time_10ths:02d}")
        else:
            raise IndexError("Time out of range ")

    def get_sealing_temperature_setpoint(self) -> None:
        return self.send_command("C")

    def get_sealing_time(self) -> None:
        return self.send_command("D")

    def get_sealing_temperature_actual(self) -> None:
        return self.send_command("F")
